
Add LoadPath "./multi_thesis".

Require Import ssreflect ssrfun ssrbool fintype bigops matrix ssrnat ssralg. 
Require Import Rstruct min_max.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.
Require Import ClassicalChoice.
Require Import topology vector metric seqRp real_matrix norm_max.
Require Import Max.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Local Scope R_scope.







Notation "v1 +^ v2" := (add_v v1 v2) (at level 50).
Notation "a *^ v" := (mult_sv a v) (at level 40).
Notation "v1 -^ v2" := (dif_v v1 v2) (at level 50).
Notation "-^ v" := (opp_v v) (at level 35).



Section funp.
Variable p: nat.

(*
Lemma Hp: (O<p)%nat.
Proof. trivial. Qed.*)

Notation Local " 'Rvec' p " := (vec R p)(at level 20): type_scope.

(*functions (Rvec p)->(Rvec p) ; we assume total functions*)

Definition add_f (f1 f2:(Rvec p)->(Rvec p)) v:= add_v (f1 v) (f2 v).
Definition dif_f (f1 f2:(Rvec p)->(Rvec p)) v:= dif_v (f1 v) (f2 v).
Definition mult_sf (x:R) (f : (Rvec p) -> (Rvec p)) v:= mult_sv x (f v).

Definition zero_f (v:(Rvec p)):= @vect0 p.

Lemma annul_f :
forall f v, mult_sf R0 f v = (@vect0 p). 
Proof.
move =>*.
by rewrite  /mult_sf /mult_sv /vect0; apply/vecP  => i; 
rewrite !vxE ; ring_simplify.
Qed.


(*function limit; we do not define the limit with respect to a set as we have total functions*)
(*we need concepts for diferent types of functions R-> Rp, Rp -> R,
Rp -> Rp*)



Lemma limit_vect0:
forall (f:(Rvec p)->(Rvec p))(v:(Rvec p)),
(forall v0, f v0 = @vect0 p) -> limit f v (@vect0 p).
Proof.
move=> f v Heq eps Heps.
exists eps.
split; [assumption|move => v0 Hdist].
rewrite (Heq v0).
have Heq0: dist_Rp (@vect0 p) (@vect0 p) = R0 by rewrite (dist_refl).
by rewrite /dist //= Heq0.
Qed.

Lemma limit_Rp_plus:
forall (f1 f2: Rvec p -> Rvec p) l1 l2 v,
limit f1 v l1 ->
limit f2 v l2 ->
limit (add_f f1 f2) v (add_v l1 l2).
Proof.
move => f1 f2 l1 l2 v Hl1 Hl2 eps Heps.
rewrite /limit in Hl1 Hl2.
elim (Hl1 (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp1 [Ha1 Hdis1].
elim (Hl2 (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp2 [Ha2 Hdis2].
exists (Rmin alp1 alp2).
split; [|move => v0 Hineq].
unfold Rmin; case (Rle_dec alp1 alp2) => H; fourier.
have Heq: dif_v (add_v (f1 v0 ) (f2 v0)) (add_v l1 l2) = add_v (dif_v (f1 v0) l1) (dif_v (f2 v0) l2).
 rewrite /dif_v /add_v; apply/vecP => i; rewrite !vxE; 
rewrite //=;ring.
rewrite /dist //= /dist_Rp Heq.
apply Rle_lt_trans with (norm (dif_v (f1 v0) l1) + (norm (dif_v (f2 v0) l2))).
apply (norm_tri ).
fold (dist_Rp (f1 v0) l1) (dist_Rp (f2 v0) l2) .
rewrite /dist //= in Hineq.
elim Hineq  => H0 Hmin.
have Hin1: R0 < dist_Rp v0 v < alp1. 
split; [assumption|].
by destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp v0 v)); fourier.
have Hin2: R0 < dist_Rp v0 v < alp2. 
split; [assumption|].
destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp v0 v)); fourier.
replace eps with (eps*/2 + eps*/2)%R; [|field].
apply Rplus_lt_compat.
exact (Hdis1 v0 Hin1).
exact (Hdis2 v0 Hin2).
Qed.

Lemma limit_R_multc:
forall (f: R -> R) k x l,
limit f x l ->
limit (fun t => k*f t) x (k*l).
Proof.
move => f k x l Hl eps He.
elim (Req_dec k R0) => Hk0.
exists eps; split; first by done.
by move => x0 Hx0; rewrite Hk0 !Rmult_0_l /dist //= /R_dist Rminus_0_r Rabs_R0.
have Hek: /Rabs k*eps > R0.
apply: Rmult_gt_0_compat ; [by apply: Rinv_0_lt_compat; apply Rabs_pos_lt|done].
elim (Hl (/Rabs k*eps) Hek) => del [Hd Hr].
exists del; split; first by done.
move => x0 Hx0.
rewrite /dist //= /R_dist -Rmult_minus_distr_l Rabs_mult.
apply (Rmult_lt_reg_l (/Rabs k)).
by apply: Rinv_0_lt_compat; apply Rabs_pos_lt.
ring_simplify.
rewrite Rinv_l; [| by apply Rgt_not_eq; apply: Rabs_pos_lt].
 rewrite Rmult_1_l.
by apply: Hr.
Qed.


Lemma limit_Rp_uniq:
forall (f:Rvec p -> Rvec p) v l l', limit f v l -> limit f v l' -> l = l'.
Proof.
move => f v l l' Hll Hll'.
rewrite -norm0_eq.
have H0: p = 0%Dn \/ (0 < p)%Dn.
case: p; [ by left | move=> n; by right].
elim: H0 => H0. 
rewrite /norm.
rewrite !big_ip_zero; try done .
have Hra: (~ (R0<dist_Rp l l')).
rewrite /not => Hdist.
elim (Hll ((dist_Rp l l')*/2) (eps2_Rgt_R0 (dist_Rp l l') Hdist)) => alp1 [Ha1 Hdis1].
elim (Hll' ((dist_Rp l l')*/2) (eps2_Rgt_R0 (dist_Rp l l') Hdist)) => alp2 [Ha2 Hdis2].
have Heq1: dist_Rp l l' = norm (dif_v l l') by done.
have Heq: forall u, dif_v l l'  = add_v (dif_v l (f u)) (dif_v (f u) l').
move => u; rewrite  /dif_v /add_v; apply/vecP =>i; rewrite !vxE;
rewrite //=; field.
have Hex: exists x, R0 < dist_Rp x v < Rmin alp1 alp2.
unfold dist_Rp.
have Hex2: R0< (Rmin alp1 alp2)*/2.
unfold Rmin; case Rle_dec; move=>*; by  apply :eps2_Rgt_R0.
 have Hex3: R0< (Rmin alp1 alp2)*/2< Rmin alp1 alp2. 
split; [done| apply Rlt_eps2_eps; fourier].
elim (ex_v_norm H0  Hex2) => ev Heq2.
exists (add_v ev v).
have Heqd: dif_v (add_v ev v) v = ev.
rewrite  /dif_v /add_v; apply/vecP =>i; rewrite !vxE  //=;
 field.
by rewrite Heqd Heq2.
elim Hex => x Hine.
have Hco: norm (add_v (dif_v l (f x)) (dif_v (f x) l')) < dist_Rp l l'.
apply Rle_lt_trans with (norm (dif_v l (f x)) + norm (dif_v (f x) l')).
apply (norm_tri ).
replace (dist_Rp l l') with (dist_Rp l l' */2 +dist_Rp l l' */2)%R;[|field]. 
elim Hine => Hmin0 Hmin1.
have Hin1: R0 < dist_Rp x v < alp1. 
split; [assumption|].
destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp x v)); fourier.
have Hin2: R0 < dist_Rp x v < alp2. 
split; [assumption|].
destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp x v)); fourier.
apply Rplus_lt_compat.
fold (dist_Rp l (f x)).
rewrite {1}dist_sym.
exact (Hdis1 x Hin1). 
fold (dist_Rp (f x) l').
exact (Hdis2 x Hin2).
rewrite -(Heq x)  /dist_Rp in Hco.
have Hco2: (~( norm (dif_v l l')< norm (dif_v l l'))); fourier.
by elim (dist_pos  l l') => H.
Qed.


(*continuity*)

(*linear functions*)

Definition linear (f: Rvec p -> (Rvec p)) :Prop:=
forall x y, add_v (f x) (f y) = f (add_v x y)/\
forall x c, mult_sv c (f x) = f (mult_sv c x).


Lemma Heine1_Rp:
forall (f: Rvec p -> Rvec p) Un x,
cont f x -> conv Un x -> conv (fun i:nat=>f (Un i)) (f x).
Proof.
move => f Un x H H0 eps0 He.
destruct (H eps0 He) as [delta [H1 H2]] .
destruct (H0 delta H1) as [N H3] .
exists N.
intros n Hn.
elim (dist_pos (Un n) x) => H4.
apply (H2 (Un n)).
exact (H3 n Hn).
have Heq: Un n = x.
by rewrite -(@dist_refl p).
by rewrite /dist  /= Heq (dist0_eq).
Qed.

Lemma same_lim:
forall xn (x:Rvec p),
(conv xn x)->
(conv (fun i:nat => xn (S i)) x).
intros.
unfold conv in H |-*.
intros.
elim (H eps H0); intros N Hxn.
exists (N-(S O))%coq_nat.
intros.
apply (Hxn (S n)).
move/leP:H1 => H1.
apply/leP.
 omega.
Qed.
 
Lemma cv_dif_Rp:
forall xn yn (x y:Rvec p) ,
conv xn x ->
conv yn y ->
conv (fun n:nat => dif_v (xn n) (yn n)) (dif_v x y).
Proof.
move => xn yn x y Hcvx Hcvy eps He.
elim (Hcvx (eps*/2)(eps2_Rgt_R0 eps He)) => N1 Hcvx1.
elim (Hcvy (eps*/2)(eps2_Rgt_R0 eps He)) => N2 Hcvy1.
exists (max N1 N2) => n Hineq.
apply Rle_lt_trans with (dist_Rp (xn n) x + dist_Rp y (yn n)).
rewrite /dist /= /dist_Rp.
replace (dif_v (dif_v (xn n) (yn n)) (dif_v x y)) with (add_v (dif_v (xn n) x) (dif_v y (yn n))).
apply (norm_tri ).
by  eq_vec.
have H1: (N1<=n)%coq_nat. 
elim (max_dec N1 N2) => H.
move/leP: Hineq => Hineq.
by rewrite H in Hineq.
move/leP: Hineq => Hineq.
by apply le_trans with (max N1 N2);[apply le_max_l|] .
have H: (N2<=n)%coq_nat.
 elim (max_dec N1 N2) => H.
move/leP : Hineq => Hineq.
by apply le_trans with (max N1 N2);[apply le_max_r|] .
move/leP: Hineq => Hineq.
by rewrite H in Hineq.
replace eps with (eps*/2+eps*/2)%R; [|field].
apply Rplus_lt_compat.
move/leP: H1 => H1.
apply (Hcvx1 n H1).
rewrite dist_sym.
move/leP: H => H2.
apply (Hcvy1 n H2).
Qed.

Lemma lim_minus_0:
forall xn x,
(conv xn x)->
(conv (fun i:nat => dif_v (xn i) (xn (S i))) (@vect0 p)).
Proof.
move => xn  x Hcv.
unfold conv => eps He .
replace (@vect0 p) with (dif_v x x).
apply cv_dif_Rp.
exact Hcv.
by apply same_lim.
exact He.
by eq_vec.
Qed.

(*some relations between dif concepts of limits
e.g. the limit in Rp is a limit on components etc. *)

Lemma echiv_lim_R_Rp:
 forall (f:R -> Rvec p)  (a: R)  (l: vec R p),
limit f a l <-> forall i, limit (fun v:R => (f v) i : R) a (l i).
Proof.

have H0: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n; by right.
elim: H0 => H0. 

move => f a l.
split=> H.
move=> i.
have H2: ~(i < 0)%Dn by done.
by elim: H2; rewrite -H0.
(*have H2: forall (u v: vec R p), u = v.
move=> u v.
apply/vecP=> i.
have H2: ~(i < 0)%Dn by done.
by elim: H2; rewrite -H0.*)
rewrite /limit /dist //= /dist_Rp /R_dist .
move=> eps Heps .
exists eps; split; first by done.
move=> x Hx.
by rewrite /norm big_ip_zero.



 split => H;
rewrite /limit /dist //= /dist_Rp /R_dist in H |-*.
move => i eps He ; elim (H eps He) => alp [Ha Hr].






exists alp; split; first by done.
move =>x Hxa.
have Hr1 := (Hr x Hxa).
apply Rle_lt_trans with (norm (dif_v (f x) l)).

replace  (Rminus (@fun_of_vec R p (f x) i) (@fun_of_vec R p l i)) with 
((dif_v (f x) l) i) ;[|by rewrite vxE].
(*
replace (f x i - l i) with ((dif_v (f x) l) i). ;[|by rewrite vxE].*)
apply: norm_max.
exact: Hr1.
move => eps He.
have Hl1: forall eps, eps > R0 -> forall i, exists alp, alp > R0 /\
            (forall x :  R,
             R0 < Rabs (x - a) < alp -> Rabs (f x i - l i)  < eps).
move=> e Hep i; exact (H i e Hep).
have Hep: eps/INR p > R0.
have Hp1: R0<INR p.
 apply lt_INR_0.
 by move/leP: H0 => Hp1.
rewrite /Rdiv.
 apply Rmult_gt_0_compat; [done| ].
by apply: Rinv_0_lt_compat; exact Hp1 .
elim (choice (fun i alp => alp > R0 /\
          (forall x : R,
           R0 < Rabs (x - a) < alp -> Rabs (f x i - l i) < eps)) (Hl1 (eps) He))=> cf Hrel.
exists (min_del cf) .
split.
apply: (min_ge0).
move => j; elim (Hrel j) => Hinq Hrest; apply: Hinq.
move =>x Hdist.
elim (norm_rew H0 (dif_v (f x) l)) => k Hk;
rewrite Hk vxE.
elim (Hrel k) => H1 H2.
apply (H2 x).
split; elim Hdist => Hd1 Hd2; first by done.
apply Rlt_le_trans with (min_del cf); [done |apply min_min].
Qed.

Lemma uniq_lim_R:
forall (f:R -> R) x l k,
limit  f x l -> 
limit f x k ->
l=k.
Proof.
move => f x l k Hl Hk.
rewrite /limit /dist //= in Hl Hk.
have H: forall eps, eps>R0-> R_dist l k <eps.
move => eps Heps.
elim (Hl (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp1 [Ha1 Hdis1].
elim (Hk (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp2 [Ha2 Hdis2].
replace eps with (eps */2 + eps */2) ; [|by field].
have Hmin: Rmin alp1 alp2 >R0 by apply Rmin_Rgt_r.
have H1: R0 < R_dist (x+ (Rmin alp1 alp2)*/2) x < alp1.
rewrite /R_dist /Rminus.
rewrite Rplus_comm -Rplus_assoc  Rplus_opp_l Rplus_0_l Rabs_right.
split; [by apply eps2_Rgt_R0|].
apply Rlt_le_trans with (Rmin alp1 alp2);[by apply Rlt_eps2_eps| by apply Rmin_l].
by rewrite /Rge; left; apply : eps2_Rgt_R0.
have H2: R0 < R_dist (x+ (Rmin alp1 alp2)*/2) x < alp2.
rewrite /R_dist /Rminus.
rewrite Rplus_comm -Rplus_assoc  Rplus_opp_l Rplus_0_l Rabs_right.
split; [by apply eps2_Rgt_R0|].
apply Rlt_le_trans with (Rmin alp1 alp2);[by apply Rlt_eps2_eps| by apply Rmin_r].
by rewrite /Rge; left; apply : eps2_Rgt_R0.
have Hd1:= Hdis1 (x + Rmin alp1 alp2 * / 2) H1.
have Hd2 := Hdis2 (x + Rmin alp1 alp2 * / 2) H2.
apply Rle_lt_trans with (R_dist (f (x + Rmin alp1 alp2 * / 2)) l + 
  R_dist (f (x + Rmin alp1 alp2 * / 2)) k).
by rewrite (R_dist_sym (f (x + Rmin alp1 alp2 * / 2)) l ) ; apply R_dist_tri.
by apply Rplus_lt_compat.
by apply cond_eq_el.
Qed.

Notation "'MR_' p" := (matrix R p p)(at level 10).
(*Notation "\vec_ ( i ) E" :=
  (@vec_of_fun p (fun i  => E)) (at level 25).
*)


Delimit Scope matrix_scope with mx.

Lemma cont_norm_matr:

forall (f: Rvec p->MR_(p)) (v:Rvec p),
(forall i, cont (fun v0:Rvec p  => \vec_(j) (f v0 i j)) v) ->
(forall eps,R0<eps-> exists del, R0<del /\
  forall v0, R0< dist_Rp v0 v < del -> norm_m ((f v0) - (f v))<eps).
Proof.
have H0: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n; by right.
elim: H0 => H0. 
move => f v Hcont eps Heps.
exists eps; split; first by done.
move=> v0 Hv0.
by rewrite /norm_m big_ip_zero.

move => f v Hcont eps Heps.
rewrite  /cont in Hcont.
have Hc1: forall eps, eps >R0 -> forall i, exists alp, alp > R0 /\
            (forall v0 :Rvec p,
             R0 < dist_Rp v0 v < alp -> 
             dist_Rp (\vec_(j) (f v0 i j)) (\vec_(j) (f v i j)) < eps).
move=> e He i.
elim (Hcont i e He) => alp [Ha Hc].
exists alp.
split; first by done.
move => v0 [H1 H2]. by apply (Hc v0).
have Hep: eps/INR p >R0.
have Hp1: R0<INR p.
by apply lt_INR_0; move/leP: H0 => Hp1.
rewrite /Rdiv.
apply Rmult_gt_0_compat.
fourier.
have Haux : R0 < / INR p.
by apply Rinv_0_lt_compat.
auto with real.
elim (choice (fun i alp => alp > R0 /\
          (forall v0 : Rvec p,
           R0 < dist_Rp v0 v < alp -> 
          dist_Rp (\vec_(j) (f v0 i j)) (\vec_(j) (f v i j)) < eps/INR p)) (Hc1 (eps/INR p) Hep))=> cf Hrel.
exists (min_del cf) .
split.
apply (min_ge0 ); move => j; elim (Hrel j) => H1 _; done.
move =>v0 Hdist.
rewrite /dist_Rp in Hrel Hdist.
elim (normM_ex  ((f (v0)) - (f v)) H0) => k Hnm.
have H1: forall x, cf x >R0.
move => x.
apply Rge_gt_trans with (min_del cf).
apply Rle_ge; apply min_min.
rewrite /Rgt.
elim Hdist => Hd1 Hd2.
fourier.
have H2:  forall x, R0 < norm (dif_v v0 v) < cf x.
move => x.
 elim Hdist => Hd1 Hd2.
split.
by done.
apply Rlt_le_trans with (min_del cf).
by done.
by apply min_min.
have H3:= (Hrel k).
elim H3 => H4 H5.
have H6:= (H5 v0 (H2 k)).
rewrite Hnm /sum_abs.
have H7: forall x , Rabs (( ( (f v0  - f v)%Ri) k) x)  < eps/INR p.
move =>x.
elim (norm_rew H0 (dif_v (\vec_(i0)(f v0 k i0)) (\vec_(i0) (f v k i0)))) => j Hrelv.
apply Rle_lt_trans with ( norm (dif_v (\vec_(i0)(f v0 k i0)) (\vec_(i0) (f v k i0)))).

have H8: (f v0 k x - f v k x)%Ri = dif_v (\vec_(i0) (f v0 k i0)) (\vec_(i0) (f v k i0)) x.
rewrite  !vxE; toR.
rewrite !mxE H8.
apply norm_max.
exact H6.
apply: (ineq_sum_el).
done.
rewrite /GRing.zero //=; fourier.
assumption.
Qed.


Lemma lema15:
forall (f: (Rvec p)->MR_(p)) (v:Rvec p) (Un:nat->Rvec p)  ,
(forall eps,R0<eps-> exists del, R0<del /\forall v0, 
  R0< dist_Rp v0 v < del -> norm_m ((f v0) - (f v))<eps) ->
conv Un v ->
(forall eps, R0<eps ->exists N, forall n, (N<=n)%coq_nat->  norm_m ((f (Un n)) - (f v))<eps).
Proof.
move =>f v Un H H0 eps0 He. 
destruct (H eps0 He) as [delta [H1 H2]] .
destruct (H0 delta H1) as [N H3] .
exists N.
intros n Hn.
elim (dist_pos  (Un n) v) => H4.
apply (H2 (Un n)).
split; first by done.
move/leP: Hn => Hn.
exact (H3 n Hn).
have Heq: Un n = v.
by rewrite -(dist_refl).
rewrite Heq.
have H5: (f v - f v)%Ri = 0%Ri.
apply/matrixP   =>i j; rewrite !mxE; toR;ring.
by rewrite H5 zeroM_0.
Qed.

Lemma cont_lim_0:
forall (f:(Rvec p)->MR_(p)) (yn xn:nat->Rvec p) (y:Rvec p),
(forall v k, cont (fun v0 :Rvec p =>(\vec_(j) (f v0 k j))) v)->
(conv xn (@vect0 p) )->
(conv yn y)->
(conv (fun i:nat=>mult_mv (f (yn i)) (xn i)) (@vect0 p)).
Proof.
move => f yn xn y Hcont Hcvx Hcvy.
have H4: norm_m (f y) = R0 \/ norm_m (f y) <> R0.
elim (Rtotal_order (norm_m (f y)) R0)=> H5; auto with real.
elim H4 => H5.
move => eps He.
have Hsqrt: R0< sqrt eps.
by apply sqrt_lt_R0.
elim (Hcvx (sqrt eps) Hsqrt) => Nx Hrel2.
have Hco: (forall eps, R0<eps ->exists N, forall n, (N<=n)%coq_nat->
    norm_m ((f (yn n)) - (f y))<eps).
apply lema15.
apply cont_norm_matr.
exact (Hcont y).
exact Hcvy.
elim (Hco (sqrt eps) Hsqrt) => Ny Hrel.
exists (max Nx Ny).
move => n Hnmax.
rewrite /dist /=  /dist_Rp dif_vect0.
have H1: mult_mv (f (yn n)) (xn n) = dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (@vect0 p)).
 rewrite mult_v0 -vecP /dif_v / @vect0 / mult_mv =>i; rewrite !vxE //=.  
toR; ring.
rewrite H1.
have H3: dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (@vect0 p)) = 
   add_v (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n))) 
  (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (@vect0 p))).
 rewrite /add_v /dif_v; apply/vecP => i; rewrite !vxE; toR; ring.
rewrite H3.
apply Rle_lt_trans with (norm (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n)))+
  norm (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (@vect0 p)))).
apply (norm_tri ).
rewrite distr_opp_mv distr_opp_vm.
apply Rle_lt_trans with (norm_m ((f (yn n)) - (f y))* norm (xn n) + 
  norm_m (f y)*norm (dif_v (xn n) (@vect0 p))).
apply Rplus_le_compat; apply (ineq_norm_mv ).
rewrite H5 Rmult_0_l Rplus_0_r. 
have H6:  sqrt eps* sqrt eps = eps.
 apply sqrt_sqrt; fourier.
rewrite -H6.
apply Rmult_ge_0_gt_0_lt_compat.
have H7:= norm_pos (xn n); fourier.
fourier.
apply Hrel.
apply le_trans with (max Nx Ny).
apply le_max_r.
by apply/leP.
rewrite /dist /= /dist_Rp in Hrel2.
replace (xn n) with (dif_v (xn n) (@vect0 p)).
apply Hrel2.
apply/leP.
apply le_trans with (max Nx Ny).
apply le_max_l.
by apply/leP.
by eq_vec.
move => eps He.
have Hpos: R0 < norm_m (f y).
have Hp1:= normM_pos (f y).
elim Hp1 => Hp2; [done|by elim H5].
have Hpos1: R0<eps/(2*norm_m (f y)).
rewrite /Rdiv Rinv_mult_distr. 
rewrite -Rmult_assoc.
repeat apply Rmult_lt_0_compat.
fourier.
fourier.
auto with real.
 apply Rgt_not_eq; fourier.
done.
have Hco: (forall eps, R0<eps ->exists N, forall n, (N<=n)%coq_nat-> 
   norm_m ( (f (yn n)) - (f y))<eps).
apply lema15.
apply cont_norm_matr.
exact (Hcont y).
exact Hcvy.
elim (Hco (norm_m (f y)) Hpos) => Nx Hrel2.
elim (Hcvx (eps/(2*norm_m (f y))) Hpos1) => Ny Hrel.
exists (max Nx Ny).
move => n Hnmax.
rewrite /dist /= /dist_Rp dif_vect0.
have H1: mult_mv (f (yn n)) (xn n) = dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (@vect0 p)).
rewrite mult_v0 -vecP /dif_v / @vect0 / mult_mv =>i; rewrite !vxE; toR ; ring.
rewrite H1.
have H3: dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (@vect0 p)) = 
  add_v (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n))) 
  (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (@vect0 p))).
by eq_vec.
rewrite H3.
apply Rle_lt_trans with (norm (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n)))+
  norm (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (@vect0 p)))).
apply (norm_tri ).
rewrite distr_opp_mv distr_opp_vm.
apply Rle_lt_trans with (norm_m ((f (yn n)) - (f y))* norm (xn n) + 
  norm_m (f y)*norm (dif_v (xn n) (@vect0 p))).
apply Rplus_le_compat; apply (ineq_norm_mv ).

have H6: norm (dif_v (xn n) (@vect0 p)) = R0 \/ norm (dif_v (xn n) (@vect0 p)) <> R0.
elim (Rtotal_order (norm (dif_v (xn n) (@vect0 p))) R0)=> H7; auto with real.
elim H6 => H7.
pattern (xn n) at 1;replace (xn n) with (dif_v (xn n) (@vect0 p)).
by rewrite H7 Rmult_0_r Rmult_0_r Rplus_0_l.
by eq_vec.
pattern (xn n) at 1;replace (xn n) with (dif_v (xn n) (@vect0 p)).
rewrite -Rmult_plus_distr_r.
replace eps with ((norm_m (f y) +norm_m (f y) )*(eps / (2*norm_m (f y)))); [|field].
apply Rmult_gt_0_lt_compat.
elim (norm_pos  (dif_v (xn n) (@vect0 p))) => H8;[done|by elim H7].
fourier.
apply  Rplus_lt_compat_r .
apply Hrel2.
apply le_trans with (max Nx Ny).
apply le_max_l.
by apply/leP. 
rewrite /dist /= /dist_Rp in Hrel.
apply Hrel.
apply/leP; apply le_trans with (max Nx Ny).
apply le_max_r.
by apply/leP.
done.
by eq_vec.
Qed.


Lemma lim_prod_minus:
forall (f:Rvec p->MR_(p)) (yn:nat->Rvec p) (y:Rvec p),
(forall v k, cont (fun v0 :Rvec p =>(\vec_(j) (f v0 k j))) v)->
(conv yn y)->
(conv (fun i:nat=>mult_mv (f (yn i)) (dif_v (yn i)  (yn (S i)))) (@vect0 p)).
Proof.
move=> f yn y Hcont Hcv.
apply cont_lim_0 with y; try done.
apply lim_minus_0 with y; done.
Qed.

End funp.
