
Add LoadPath "./multi_thesis".

Require Import ssreflect ssrbool ssrnat eqtype fintype bigops matrix ssralg.
Require Import Rstruct min_max topology  vector metric seqRp funRp 
   real_matrix norm_max deriv.
Require Import   Reals Fourier.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Local Scope ring_scope.

Open Local Scope R_scope.

Delimit Scope ring_scope with Ri.

Section Kantorovitch.

Import GRing.Theory.

Variable p': nat.
Notation p := p'.+1.

Lemma Hp: (O<p)%Dn.
by done. Qed.


Notation Local " 'Rvec' p " := (vec R p)(at level 20): type_scope.
Notation "v1 +^ v2" := (add_v v1 v2) (at level 50).
Notation "a *^ v" := (mult_sv a v) (at level 40).
Notation "v1 -^ v2" := (dif_v v1 v2) (at level 50).
Notation "-^ v" := (opp_v v) (at level 35).
Notation "'MR_' p" := (matrix R p p)(at level 10).
(*Notation "A -m B" := (dif_m A B) (at level 50, format "A  -m  B").*)
Notation "A ^-1m" := (invmx  A) (at level 10, format "A ^-1m") .

Variable a0:R.
Variable b0:R.
Variable c:R.
Variable keps:R.
Variable X0:Rvec p.
Variable f:(Rvec p)->Rvec p.


Fixpoint An (n:nat) {struct n} : R :=
match n with
|0=>a0
| S n => 2*(An n)
end.

Fixpoint Bn (n:nat) : R:=
match n with
|0=>b0
|S n => INR p*(An n)*(Bn n)*(Bn n)*c
end.

Fixpoint En (n:nat): R :=
match n with
|0 => keps
|S n => (En n)/2
end.

Hypothesis Hkepos: 0<keps.
(*
Hypothesis H_a0:0<a0.
Hypothesis H_b0:0<b0.
Hypothesis H_c:0<c.
*)

Hypothesis A0_b0_c: 2*(INR p)*a0*b0*c<=1.

Definition c_disc_Rp (x:Rvec p) (delta : R)(y:Rvec p) : Prop :=
0<delta/\(dist_Rp x y)<=delta.


Hypothesis Hcontf:
forall v, cont  f v.
(* prove that a diff fun is continuous!*)

Hypothesis pr: forall (x : Rvec p) (x0 : 'I_p), dbl_pt  f x0 x. (* once derivable*)
Hypothesis pr2: forall (i j:'I_p) (a:Rvec p), dbl_pt_2  j a (fun x => pr x i) . (*twice der*) 
Hypothesis der_cont1: forall a i, cont  (fun x => der_p (pr x i)) a. (*cont 1st der*)
Hypothesis der_cont2: forall a i j, cont  
  (fun v => der_p_2 (pr2 i j v)) a .
(*cont 2nd der*)

(*this is unnecessary... but it simplifies stuff;*)
Hypothesis der_bo: forall (v0:Rvec p) (i j : 'I_p),
\sum_( k < p) Rabs (der_p_2  (pr2 j k v0) i) <=c.

(*
Hypothesis pr: forall a i, dbl_pt p Hp f i a.
*)
Variable f': Rvec p -> MR_(p).
Hypothesis Hjac: forall a, f' a   =  Jac (pr a).


(*Definition f' (a:Rvec p):=
Jac p  f a (pr a).
*)

Hypothesis Hinv0 : \det (f' X0) <> 0.
Hypothesis Ha: norm_m ((f' X0)^-1m) <= a0.
Hypothesis Hb: norm (mult_mv ( (f' X0)^-1m) (f X0)) <= b0.
Hypothesis Hb0: b0<=keps*/2.



Lemma Hder_bo:
forall x y, norm_m ((f' x) - (f' y)) <=c*(INR p)*norm (dif_v x y).
Proof.
move=> x y.

have H3: forall f g i a pr1 pr2, (forall t , f t = g t) -> @dp_1 p f i a pr1 = @dp_1 p g i a pr2.
move=> h g ind av prh prg Hhg.
rewrite /dp_1.
case prh => lh Hlh; rewrite //=.
case prg => lg Hlg; rewrite //=.
rewrite /part_deriv_pt_1 in Hlh Hlg.
have H1: limit (fun t : Base R_met => (g (av +^ t *^ base_v ind) - g av) / t)
        0 lh.
red; apply echiv_lim_f1 with (fun t : Base R_met => 
   ((h (av +^ t *^ base_v ind) - h av) / t):Base R_met);
last by done .
by move=> t;  rewrite !Hhg.
by apply uniq_lim_R with 
(fun t : Base R_met => ((g (av +^ t *^ base_v ind) - g av) / t) : Base R_met) R0.
rewrite Hjac /Jac.
have Heq1: x = (y+^(x-^y)).
 rewrite -vecP => i; rewrite !vxE //=; ring.
(*by rewrite /Rminus  Rplus_comm Rplus_assoc Rplus_opp_l Rplus_0_r.*)
rewrite {1 2}Heq1.
rewrite Rmult_assoc Rmult_comm.

have Hl1:= (@lemme1 p (fun i j y=> @der_p p  f j y (pr y j) i ) y (x-^y) c).
have pr0 : forall (x x0 x1 : 'I_p) (x2 : Rvec p),
              @dbl_pt_1 p (fun y : Rvec p => @der_p p  f x0 y (pr y x0) x) x1
                x2.
move=>  k j i v.
have Hedp:= @eq_der_p p  f j v (pr v j) k .
have Heq2: forall v, @der_p p  f j v (pr v j) k =
       @dp_1 p (fun v : Rvec p => f v k) j v
         (@echiv_der_p2 p  f j v (pr v j) k).
 by move=> v0; apply eq_der_p.
have hdbl2: @dbl_pt_1 p (fun v => @dp_1 p (fun v0 : Rvec p => f v0 k) j v
         (@echiv_der_p2 p  f j v (pr v j) k)) i v.
apply dbl_2_impl_dp_dp_1.

have H31: @dbl_pt_2 p  f j i v (fun v => pr v j).
apply pr2.
have H4:= @echiv_der_p2_gen p  f j i v (fun v=> pr v j) H31 k.
elim H4 => dp2 Hdp2; exists dp2.
red in Hdp2 |-*; red in Hdp2 |-*; red in Hdp2 |-*.
apply echiv_lim_f1   with (fun t : Base R_met =>
          (@dp_1 p (fun v : Rvec p => f v k) j (v +^ t *^ base_v i)
             (@echiv_der_p3 p  f j k (fun v : Rvec p => pr v j)
                (v +^ t *^ base_v i)) -
           @dp_1 p (fun v : Rvec p => f v k) j v
             (@echiv_der_p3 p  f j k (fun v : Rvec p => pr v j) v)) / t); last by done.
move => t.
have heq1: @dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
   (@echiv_der_p3 p  f j k (fun v0 : Rvec p => pr v0 j) (v +^ t *^ base_v i)) =
(@dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
   (@echiv_der_p2 p  f j (v +^ t *^ base_v i) (pr (v +^ t *^ base_v i) j) k)).
apply dp_1_pr_nu.
have heq2: @dp_1 p (fun v0 : Rvec p => f v0 k) j v
   (@echiv_der_p3 p  f j k (fun v0 : Rvec p => pr v0 j) v)  =
@dp_1 p (fun v0 : Rvec p => f v0 k) j v (@echiv_der_p2 p  f j v (pr v j) k) .
apply dp_1_pr_nu.
by rewrite heq1 heq2.
elim hdbl2 => dp2 Hdp2.
exists dp2.
red in Hdp2 |-*; red in Hdp2 |-*.
apply echiv_lim_f1 with  (fun t : Base R_met =>
          (@dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
             (@echiv_der_p2 p  f j (v +^ t *^ base_v i)
                (pr (v +^ t *^ base_v i) j) k) -
           @dp_1 p (fun v0 : Rvec p => f v0 k) j v
             (@echiv_der_p2 p  f j v (pr v j) k)) / t); last by done.
move => t.
have heq1: (@dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
   (@echiv_der_p2 p  f j (v +^ t *^ base_v i) (pr (v +^ t *^ base_v i) j) k))=
(@der_p p  f j (v +^ t *^ base_v i) (pr (v +^ t *^ base_v i) j) k).
by rewrite Heq2.
have heq2:  @dp_1 p (fun v0 : Rvec p => f v0 k) j v (@echiv_der_p2 p  f j v (pr v j) k)=
@der_p p  f j v (pr v j) k.
by rewrite Heq2.
by rewrite heq1 heq2.
have Hl2:= Hl1 pr0.
have Hcont1: forall (i j i0 : 'I_p) (v0 : Rvec p),
       cont 
         (fun x : Rvec p =>
          @dp_1 p (fun y : Rvec p => @der_p p  f j y (pr y j) i) i0 x
            (pr0 i j i0 x)) v0.
move=> j i k v.

apply cont_eq_fun with (fun v => (@der_p_2 p  f i k v (fun x => pr x i) (pr2 i k v) j)).
move=> x0.
rewrite dp_12_der_p2_eq dp_2_impl_dp_1.

apply H3.
move=> w.
rewrite eq_der_p.
apply H3.
move=> t; done.
apply: rel_cont_Rp.
apply der_cont2.
have Hl3:= Hl2 Hcont1.
have Hbound:  forall (x : Rvec p) (i j : 'I_p),
\sum_(k<p) Rabs  (@dp_1 p (fun y : Rvec p => 
  @der_p p  f j y (pr y j) i) k x (pr0 i j k x)) <= c.
move => v0 i j.
apply Rle_trans with (\sum_( k < p) Rabs 
  (@der_p_2 p  f j k v0 (fun v => pr v j) (pr2 j k v0) i) ).
right.
apply: eq_bigr => k _.
f_equal.
rewrite dp_12_der_p2_eq dp_2_impl_dp_1.
apply H3.
move=> t.
rewrite eq_der_p.
apply H3.
move=> w; done.
apply der_bo.
rewrite Hjac; apply (Hl3 Hbound).
Qed.




Fixpoint Xn (n:nat): Rvec p:=
match n with 
|0 => X0
|S n =>dif_v (Xn n) (mult_mv ((f' (Xn n))^-1m) (f (Xn n)))
end . 



Lemma ineq12:
forall (x y:Rvec p),
norm (dif_v (f x) (add_v (f y)  (mult_mv  (f' y) (dif_v x y))))
  <=1/2*INR p*Rsqr (norm (dif_v x y))*c.
Proof.
move => x y.
have Heqy: x = y +^ (x -^ y).
rewrite -vecP => i; rewrite !vxE /=; ring.
(*by rewrite /Rminus  Rplus_comm Rplus_assoc Rplus_opp_l Rplus_0_r.*)
rewrite {1}Heqy.
have Heq: (f (y +^ (x -^ y)) -^ (f y +^ mult_mv (f' y) (x -^ y)))=
  ((f (y +^ (x -^ y)) -^ f y) -^ mult_mv (f' y) (x -^ y)).
by apply/vecP => i; rewrite !vxE; ring_simplify.
rewrite Heq.
by rewrite Hjac; apply lemme2 with pr2.
Qed.

Lemma keps_pos:
forall n, 0<En n.
Proof.
induction n.
simpl; auto.
simpl; fourier.
Qed.

Lemma inc_neigh_Rp:
forall x y r, 0<r -> dist_Rp x y <= r*/2 ->
(forall v, c_disc_Rp y (r*/2) v -> c_disc_Rp x r v).
Proof.
move => x y r Hr Hdist v [He0 Hdisc].
unfold c_disc_Rp; split; first by done.
apply Rle_trans with (dist_Rp x y + dist_Rp y v).
 apply vector.dist_tri; exact Hp.
rewrite -(eps2 r); fourier.
Qed.

Lemma inc_neigh_ech:
forall (x y:Rvec p) r, 0<r -> dist_Rp x y <= r*/2 ->
(forall v, dist_Rp y v<=(r*/2)  -> dist_Rp x v <=r).
Proof.
move => x y r Hr H1 v H2.
apply Rle_trans with (dist_Rp x y + dist_Rp y v).
by apply vector.dist_tri; exact Hp.
rewrite -(eps2 r); fourier.
Qed.


Lemma Rinv_ineq:
forall a b,0<=a->0<=b->
1<=a*b -> 1/b<=a.
Proof.
move =>a b H1 H2 H3.
unfold Rdiv.
have H4: 0<b.
have H5: 0<a*b.
apply Rlt_le_trans with 1; fourier.
have Hco: ~ (b<=0).
unfold not => Hco1.
have Hco4: b=0. 
by apply Rle_antisym.
have Hco2: 0=a*b. by rewrite Hco4; ring.
have Hco3: ~(0=a*b).
by apply Rlt_not_eq.
contradiction.
elim (Rle_or_lt b 0)=>H.
contradiction.
done.
replace a with (a*b*/b); [|field].
apply Rmult_le_compat_r.
unfold Rle; left.
auto with real.
assumption.
apply Rgt_not_eq.
fourier.
Qed.


Lemma norm_m_inv:
forall a:'M_p, (a^-1m) *m a = 1%Ri  ->
 1/(norm_m a) <= norm_m (a^-1m).
Proof.
move => a H.
have H2: 1 <= norm_m ( a^-1m)*norm_m a.
have H1: @norm_m p p 1 = 1.
apply (norm_unitM ).
rewrite -H1 -H.
apply (ineq_norm_m ).
apply Rinv_ineq.
apply normM_pos.
apply normM_pos.
assumption.
Qed.

Lemma a0_ge_0:
0<=a0.
Proof.
apply Rle_trans with (norm_m ((f' X0)^-1m));[
apply normM_pos|done].
Qed.

Lemma b0_ge_0:
0<=b0.
Proof.
apply Rle_trans with (norm (mult_mv ( (f' X0)^-1m) (f X0))).
 apply norm_pos; exact Hp.
done.
Qed.

Lemma c_ge_0:
0<=c.
Proof.
apply Rle_trans with (
\sum_( k < p) Rabs (@der_p_2  p f (I0 Hp ) k X0 
  (fun v => pr v (I0 Hp )) (pr2 (I0 Hp ) k X0) (I0 Hp ))).
apply Rle_trans  with 
(Rabs (\sum_( k < p) (@der_p_2 p  f (I0 Hp ) k X0 
  (fun v => pr v (I0 Hp )) (pr2 (I0 Hp) k X0) (I0 Hp)))).
apply Rabs_pos.
by apply: abs_sum_le_f.
done.
Qed.



Lemma An_gt_0:
forall n, 0<=An n.
Proof.
induction n.
simpl; auto.
apply a0_ge_0.
simpl; apply Rmult_le_pos; [fourier|auto].
Qed.

Lemma An_bn_c0:
forall (n:nat) , 0<=2*INR p*(An n)*(Bn n)*c.
Proof.
induction n.
rewrite /An /Bn .
 repeat apply Rmult_le_pos. 
 fourier.
 apply pos_INR.
apply a0_ge_0.
apply b0_ge_0.
apply c_ge_0.
simpl.
set (ip:= match p' with
    | 0%Dn => 1
    | _.+1 => INR p' + 1
    end).
have Hd: ip = INR p by done.
rewrite Hd.
replace (2 * INR p * (2 * An n) * (INR p*An n * Bn n * Bn n * c) * c) 
  with ((2*INR p*An n*Bn n * c)*(2*INR p*An n*Bn n * c));
 [idtac | ring].
fold (Rsqr (2*INR p*An n*Bn n*c)).
auto with real.
Qed.

Lemma An_bn_c1:
forall (n:nat) , 2*(INR p)*(An n)*(Bn n)*c<=1.
Proof.
induction n.
simpl; auto.
simpl.
set (ip:= match p' with
    | 0%Dn => 1
    | _.+1 => INR p' + 1
    end).
have Hd: ip = INR p by done.
rewrite Hd.
replace (2 * INR p*(2 * An n) * (INR p*An n * Bn n * Bn n * c) * c) 
  with ((2*INR p*An n*Bn n * c)*(2*INR p*An n*Bn n * c));
 [idtac | ring].
fold (Rsqr (2*INR p*An n*Bn n*c)).
pattern 1 at 3;rewrite <- Rsqr_1.
apply Rsqr_incr_1.
done.
apply An_bn_c0.
fourier.
Qed.

Lemma Bn_gt_0:
forall n, 0<=Bn n.
Proof.
induction n.
simpl; auto.
apply b0_ge_0.
simpl; repeat apply Rmult_le_pos; auto.
rewrite -/(INR p).
by apply pos_INR.
by apply An_gt_0.
apply c_ge_0.
Qed.

Lemma Bn_en_le:
forall n,
Bn ( n) <= En ( n) * / 2.
Proof.
induction n; rewrite //=.
rewrite -/(INR p).
replace (INR p * An n * Bn n * Bn n * c) with 
  (Bn n*(2*INR p*An n* Bn n*c)*/2); [|field].

apply Rmult_le_compat_r.
fourier.
replace (En n/2) with (En n/2*1);[|field].
apply Rmult_le_compat.
 apply Bn_gt_0. 
 apply An_bn_c0.
done.
apply An_bn_c1.
Qed.

Lemma bad_four:
forall r,0<=r -> r<=1/2 ->
1/(1-r) <=2.
Proof.
move => r H1 H2.
have H3:1/2 <=  1-r .
fourier.
apply Rinv_ineq.
fourier.
fourier.
pattern 1 at 1;replace 1 with (2*(1/2)); [|field].
apply Rmult_le_compat; fourier.
Qed.


Lemma about_seq:
forall n, 
(n<>0%nat-> forall v, 
   (dist_Rp (Xn n) v <= (En n))->
   (dist_Rp (Xn (Peano.pred n)) v <= (En (Peano.pred n)))) /\
\det (f' (Xn n)) <>0/\
norm_m ( (f' (Xn n))^-1m) <= An n/\
norm (mult_mv ((f' (Xn n))^-1m) (f (Xn n)))<= Bn n/\
Bn n<= (En n)*/2.
Proof.
induction n.
rewrite //=.
elim IHn => [IHdisc [IHinv [IHa [IHb1 IHb2]]]].
have  Hinvx0:  \det (f' (Xn n)) != 0.
case X: (\det (f' (Xn n)) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := IHinv X.
done.
split.
move => Hdif v Hdisc.
rewrite //= .
apply inc_neigh_ech with (Xn (S n)).
 by apply keps_pos.
rewrite /dist_Rp.
have Heq: dif_v (Xn n)  (Xn (S n)) = mult_mv ((f' (Xn n))^-1m) (f (Xn n)).
by apply/vecP =>i; rewrite vxE //= vxE /=; ring_simplify.
rewrite Heq; by apply Rle_trans with (Bn n).
by rewrite //En in Hdisc.
have Hinv: ((f' (Xn n))^-1m * (f' (Xn n)) = 1)%Ri.
apply: mulVmx.
done.
have Ha3: norm_m ( 1 - (( (f' (Xn n))^-1m) * (f' (Xn (S n)))))%Ri <= 1/2.
rewrite -Hinv.
rewrite -mulr_subr.

apply Rle_trans with ((norm_m ((f' (Xn n))^-1m))*
 norm_m (( (f' (Xn n)) - ((f' (Xn (S n))))))).
apply: (ineq_norm_m ).
apply Rle_trans with ((An n)*norm_m ( (f' (Xn n)) - (f' (Xn (S n))))).
apply Rmult_le_compat.
elim (normM_ex  ((f' (Xn n))^-1m) Hp) => k Heq.
rewrite Heq.
apply (sum_abs_pos ).
elim (normM_ex  ( (f' (Xn n)) - (f' (Xn (S n)))) Hp) => k Heq.
rewrite Heq.
apply (sum_abs_pos ).
done.
by unfold Rle; right.
apply Rle_trans with ((An n)*((INR p)*(Bn n)*c)).
apply Rmult_le_compat_l.
apply An_gt_0.
have Ha2: norm (dif_v (Xn n) (Xn (S n))) <= Bn n.
have Heq2: dif_v (Xn n) (Xn (S n)) = mult_mv ( (f' (Xn n))^-1m) (f (Xn n)).
rewrite //=.
by apply/vecP =>i; rewrite vxE //= vxE /=; ring_simplify.
by rewrite Heq2.
replace (INR p*Bn n*c) with (c*INR p*Bn n);[| ring].
apply Rle_trans with (c*INR p* norm (dif_v (Xn n) (Xn (S n)))).
apply Hder_bo.
apply Rmult_le_compat_l.
apply Rmult_le_pos.
exact c_ge_0.
 apply pos_INR.
exact Ha2.
replace  (An n * (INR p * Bn n * c)) with (1/2*(2*INR p*An n*Bn n*c)); [|field].
pattern (1/2) at 2; replace (1/2) with (1/2*1); [|field].
apply Rmult_le_compat_l.
 fourier.
apply An_bn_c1.
have Ha2: \det ((f' (Xn n))^-1m *m (f' (Xn (S n)))) <> 0.
apply matr_inv_norm.
apply Rle_lt_trans with (1/2);[done|].
rewrite /GRing.one //=.
fourier.
rewrite detM in Ha2.
have Hinvs: \det (f' (Xn (S n))) <> 0.
rewrite /GRing.mul in Ha2.
by elim (RIneq.Rmult_neq_0_reg  _ _ Ha2) .
split; first by done.
have Ha1: ((f' (Xn (S n)))^-1 = 
  (( ((f' (Xn n))^-1) * (f' (Xn (S n))))^-1)%Ri * ((f' (Xn n))^-1))%Ri.
 rewrite invr_mul.
rewrite  -mulrA invrK divrr.
by rewrite mulr1.
done.
by rewrite unitr_inv.
by move/eqP: Hinvs => Hinvs.
(*done.
rewrite mulVmx .
rewrite  mulmx1; done.
done.
done.
apply neq0_Dmx; done.
case X: (\det (f' (Xn (S n))) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := Hinvs X.
done.*)
have Ha4:  norm_m ((f' (Xn (S n))^-1m)) <= An (S n).
move: Ha1; toR => Ha1.
rewrite Ha1.
apply Rle_trans with (norm_m (((f' (Xn n))^-1m *m f' (Xn (S n)))^-1m) * 
  norm_m ((f' (Xn n))^-1m)).
apply (ineq_norm_m  ).
simpl An.
apply Rmult_le_compat .
apply normM_pos.
apply normM_pos.
replace ( ( (f' (Xn n))^-1m) *m (f' (Xn (S n)))) with 
       ( 1 - ( 1 - (( (f' (Xn n))^-1m) *m (f' (Xn (S n))))))%Ri.
apply Rle_trans with (1/(1-( norm_m ( 1 - ( ((f' (Xn n))^-1m) *m (f' (Xn (S n)))))))).
apply: norm_inv_ser.
apply Rle_lt_trans with (1/2);[done|].
rewrite /GRing.one //=. fourier.
apply bad_four.
apply normM_pos.
done.
toR; rewrite -matrixP => i j;
by rewrite !mxE ; toR;  ring_simplify. 
by elim IHn => [_ [_ [H _]]].
split; first by done.
split.
simpl Bn.
apply Rle_trans with (norm_m ( (f' (Xn (S n)))^-1m)* norm (f (Xn (S n)))).
apply (ineq_norm_mv ).
rewrite -/(INR p).
replace (INR p * An n * Bn n * Bn n * c) with (2*An n*(1/2*INR p* Bn n*Bn n*c)); [|field].
apply Rmult_le_compat.
apply normM_pos.
apply (norm_pos ).
by simpl An in Ha4.
have Hxn: f (Xn (S n)) = (dif_v (f (Xn (S n))) (add_v (f (Xn n)) 
   (mult_mv (f'(Xn  n )) (dif_v (Xn (S n)) (Xn n))))).
have Hxna: (add_v (f (Xn n))  (mult_mv (f'(Xn  n )) (dif_v (Xn (S n)) (Xn n)))) = @vect0 p.
simpl Xn.
replace (dif_v (dif_v (Xn n) (mult_mv ( (f' (Xn n))^-1m) (f (Xn n)))) (Xn n))
 with (mult_mv ( (f' (Xn n))^-1m) (opp_v (f (Xn n)))).
apply/vecP => i; rewrite /dif_v /vect0 vxE.
rewrite mult_mv_assoc .
replace  ( (f' (Xn n) *m (f' (Xn n))^-1m)) with ( (f' (Xn n) * (f' (Xn n))^-1))%Ri.
rewrite mulrV .
rewrite unitM_el_neu.
by rewrite  !vxE /=; ring.
done.
by toR.
rewrite opp_distr; apply/vecP => i; rewrite /dif_v /opp_v !vxE /=; ring.
rewrite Hxna -vecP /vect0 /dif_v =>i ; rewrite !vxE /=; ring.
auto with real.
rewrite Hxn.
apply Rle_trans with (1/2 *INR p* (Rsqr (norm (dif_v (Xn (S n)) (Xn n))))*c).
apply ineq12.
apply Rmult_le_compat_r.
apply c_ge_0.
replace (1/2*INR p*Bn n*Bn n) with ((1/2*INR p)*(Bn n*Bn n)); [|ring].
apply Rmult_le_compat_l.
unfold Rle; left; apply Rmult_lt_0_compat.
fourier.
by apply lt_INR_0; apply/leP.
fold (Rsqr (Bn n)).
fold (dist_Rp (Xn (S n)) (Xn n)).
rewrite vector.dist_sym.
apply Rsqr_incr_1.
rewrite /dist_Rp.
replace (dif_v (Xn  n) (Xn (S n))) with (mult_mv ( (f' (Xn n))^-1m) (f (Xn n))).
by elim IHn => [_[_[_[H _]]]].
simpl Xn.
 apply/vecP =>i.
toR.
rewrite !vxE /=; ring.  
apply (vector.dist_pos ).
apply Bn_gt_0.
apply Bn_en_le.
Qed.

Lemma Cauchy_crit_Rp2:
forall Un: nat -> Rvec p,
isCauchy Un <->
(forall eps0:R,
eps0>0->
exists N:nat,
(forall n p:nat,
(n>=N)%coq_nat->dist_Rp (Un n) (Un (n+p)%coq_nat) < eps0)).
Proof.
move =>Un.
split => H.
unfold isCauchy in H.
move => eps0 H0.
elim (H eps0 H0); intros N Hchy.
exists N.
move => n p0 Hineq.
apply (Hchy n (n+p0)%coq_nat).
apply/leP; omega.
apply/leP; omega.
unfold isCauchy.
move =>eps0 H0.
elim (H eps0 H0)=> N Hcrit.
exists N.
move => n m Hineq.
elim (le_or_lt n m); intros.
replace m with (n+(m-n)%coq_nat)%coq_nat;[| omega].
apply (Hcrit n (m-n)%coq_nat).
apply/leP;assumption.
rewrite /dist /= vector.dist_sym .
replace n with (m+(n - m)%coq_nat)%coq_nat;[|omega].
apply (Hcrit m (n - m)%coq_nat).
apply/leP; assumption.
Qed.

Lemma disc_incl_discRp:
( forall n : nat,  forall v : Rvec p,
         dist_Rp (Xn (S n)) v <= En (S n) -> dist_Rp (Xn n) v <= En  n) ->
(forall (n p0:nat), forall v, dist_Rp (Xn (n+p0)%coq_nat) v <= 
  En (n+p0)%coq_nat -> dist_Rp (Xn n) v <= En  n).
Proof.
move => H n p0 v H1.
induction p0.
by replace (n+0)%coq_nat with n in H1; [|omega].
replace (n+S p0)%coq_nat with (S (n+p0)%coq_nat) in H1;[|omega].
have H2: dist_Rp (Xn (n + p0)%coq_nat) v <= En (n + p0)%coq_nat.
apply (H (n+p0)%coq_nat v H1).
exact (IHp0 H2).
Qed.

Lemma norm_eq:
forall v:Rvec p,
dist_Rp v v = 0.
Proof.
by move => v; rewrite /dist_Rp norm0_eq. Qed.

Lemma disc_incl_pct:
( forall n : nat,  forall v : Rvec p,
         dist_Rp (Xn (S n)) v <= En (S n) -> dist_Rp (Xn n) v <= En  n) ->
(forall n p0, dist_Rp (Xn n) (Xn (n+p0)%nat) <= En n).
Proof.
move => H1 n p0 .
have H0:= (disc_incl_discRp H1). 
have H3:= H0 n p0.
apply (H3 (Xn (n+p0)%nat)).
rewrite norm_eq.
unfold Rle; left; apply keps_pos.
Qed.

(*borrowed and new proofs to apply archimede's principle*)
Theorem INR_Zabs_nat_eq :
  forall (z:Z), (0 <= z)%Z -> (IZR z = INR(Zabs_nat z))%R.
unfold Zabs_nat.
intros z; case z.
simpl.
intros; auto.
unfold IZR.
intros; auto.
compute.
intros x H'; elim H'; auto.
Qed.

Lemma pow2_larger :
  forall n, (INR n <= 2^n)%R.
induction n.
simpl; fourier.
replace (S n) with (1+n)%coq_nat.
rewrite pow_add; rewrite pow_1.
assert (Hlem: (forall a b, 1<= b -> a<=b -> 1+a<= 2*b)%R).
intros; fourier.
rewrite plus_INR; simpl (INR 1).
apply Hlem; auto.
apply pow_R1_Rle;fourier.
omega.
Qed.

Lemma pow_half_smaller :
 forall r, (0 < r) -> (1/(2^(Zabs_nat(up (1/r)))) <= r).
intros r Hpos.
elim (archimed (1/r)); intros Harc1 Harc2.
apply (Rmult_le_reg_l (2^(Zabs_nat (up (1/r))))).
apply pow_lt.
fourier.
replace (2 ^ Zabs_nat (up (1 / r)) * (1 / 2 ^ Zabs_nat (up (1 / r))))%R
 with 1%R.
apply Rle_trans with ((INR (Zabs_nat (up (1/r))))*r)%R.
rewrite <- INR_Zabs_nat_eq.
pattern 1%R at 1; replace 1%R with (1/r * r)%R;[| field].
apply Rmult_le_compat_r.
fourier.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
by apply Rgt_not_eq.
apply le_IZR.
apply Rle_trans with (1/r)%R.
replace (1/r)%R with (/r)%R.
simpl IZR.
assert (0 < / r)%R.
apply Rinv_0_lt_compat; auto.
unfold Rle; left; assumption.
field.
apply Rgt_not_eq; auto.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
apply Rmult_le_compat_r.
fourier.
apply pow2_larger.
field.
apply pow_nonzero.
discrR.
Qed.

Lemma archim:
forall v, 0 < v -> exists n:nat, (1/(2^n) <= v)%R.
intros v Hvpos.
exists (Zabs_nat (up (1/v)))%nat.
destruct (archimed (1/v)) as [Hb1 _].
apply pow_half_smaller.
assumption.
Qed.

Lemma archim_lt:
(forall v, 0 < v -> exists n:nat, 1/(2^n) <= v/2).
intros v Hvpos.
exists (Zabs_nat (up (1/v))+1)%nat.
destruct (archimed (1/v)) as [Hb1 _].
rewrite pow_add.
replace (1 / (2 ^ Zabs_nat (up (1 / v))*2^1))%R
 with ((1 / (2 ^ Zabs_nat (up (1 / v))))*/2)%R.
replace (v/2)%R with (v*/2)%R.
apply Rmult_le_compat_r.
assert (0 < /2)%R.
apply Rinv_0_lt_compat;auto.
fourier.
fourier.
apply pow_half_smaller.
fourier.
field.
replace (2^1)%R with 2%R.
field.
assert (2^Zabs_nat(up(1*/v)) <> 0)%R.
apply pow_nonzero;auto.
discrR.
simpl;assumption.
simpl;field.
Qed.

Lemma archim_lt2:
forall v, 0 < v -> exists n:nat, (1/(2^n) < v)%R.
intros v Hvpos.
assert (Hhalf: exists n:nat, 1/(2^n) <= v/2).
apply archim_lt; assumption.
elim Hhalf; intros n Hineq.
exists n.
assert (v/2<v).
fourier.
unfold Rle in Hineq; elim Hineq; intros.
apply Rlt_trans with (v/2); assumption.
rewrite H0.
assumption.
Qed.

Lemma pow2_diff0:
forall n:nat,
2^n>0.
induction n; simpl; fourier.
Qed.

Lemma archim2_lt:
forall v k, 0 < v ->0<k-> exists n:nat, (k/(2^n) < v)%R.
intros v k Hvpos Hkpos.
assert (exists n:nat, 1/2^n<v/k).
apply (archim_lt2 ).
unfold Rdiv.
apply Rlt_mult_inv_pos; assumption.
elim H; intros N Hprop.
exists N.
replace (k/2^N) with (k*(1/2^N));[| field].
replace v with (k*(v/k)).
apply (Rmult_lt_compat_l k (1/2^N) (v/k)).
fourier.
assumption.
field.
apply Rgt_not_eq.
fourier.
replace (/2^N) with (1/2^N);[| field].
trivial.
apply Rgt_not_eq.
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.


Lemma archim3_lt:
forall v k, 0 < v ->0<=k-> exists n:nat, (k/(2^n) < v)%R.
intros v k Hvpos Hkpos.
elim Hkpos; intros Heq.
assert (exists n:nat, 1/2^n<v/k).
apply (archim_lt2 ).
unfold Rdiv.
apply Rlt_mult_inv_pos; assumption.
elim H; intros N Hprop.
exists N.
replace (k/2^N) with (k*(1/2^N));[| field].
replace v with (k*(v/k)).
apply (Rmult_lt_compat_l k (1/2^N) (v/k)).
fourier.
assumption.
field.
apply Rgt_not_eq.
fourier.
replace (/2^N) with (1/2^N);[| field].
trivial.
apply Rgt_not_eq.
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
exists O; rewrite -Heq //=; fourier.
Qed.

Lemma pow2_monot:
forall (n :nat), keps/2^(S n)<=keps/2^n.
Proof.
induction n .
simpl.
fourier.
simpl in IHn |-*.
replace (keps/(2*(2*2^n))) with (1/2*(keps/(2*2^n))) ;
  [|field; auto;apply Rgt_not_eq; apply pow2_diff0].
pattern (keps/(2*2^n)) at 2;replace (keps/(2*2^n)) with (1/2*(keps/2^n));
  [| field; apply Rgt_not_eq; apply pow2_diff0].
apply (Rmult_le_compat_l (1/2) (keps/(2*2^n)) (keps/2^n)).
fourier.
assumption.
Qed.

Lemma pow2_monot2:
forall (n p:nat), 
keps/2^(n+p)%coq_nat<=keps/2^n.
intros n p0.
induction p0.
replace (n+0)%coq_nat with n; [| omega].
unfold Rle; right; trivial.
replace (n+S p0 )%coq_nat with (S (n+p0)%coq_nat);[| omega].
assert (IH2: keps/2^S (n+p0)%nat<=keps/2^(n+p0)%nat).
apply pow2_monot.
apply Rle_trans with (keps/2^(n+p0)%nat); assumption.
Qed.

Lemma pow2_monot3:
forall (n N:nat), (n>=N)%coq_nat->
keps/2^n<=keps/2^N.
intros.
replace n with (N+(n-N)%coq_nat)%coq_nat;[| omega].
apply (pow2_monot2 N (n-N)%coq_nat).
Qed.

Lemma transf_en:
forall n:nat,
En n = keps/2^n.
Proof.
induction n.
simpl;field.
simpl En.
rewrite IHn.
simpl; field.
apply Rgt_not_eq.
apply pow2_diff0.
Qed.

Lemma An_bn_c12:
forall (n:nat) , INR p*(An n)*(Bn n)*Bn n*c<=1/2*Bn n.
Proof.
intros.
assert ( 2*INR p*(An n)*(Bn n)*c<=1).
apply An_bn_c1.
assert  (INR p*(An n)*(Bn n)*c<=1/2).
unfold Rdiv.
replace (INR p*An n*Bn n*c) with ((1/2)*(2*INR p*(An n)*(Bn n)*c));[| field].
replace (1*/2) with (1/2*1); [|field].
apply (Rmult_le_compat_l (1/2) (2*INR p* (An n)*(Bn n)*c) 1); fourier .
replace (INR p*(An n)*(Bn n)*Bn n*c) with (INR p*An n*Bn n*c*Bn n);[| ring].
apply Rmult_le_compat_r ; [|assumption].
apply Bn_gt_0;assumption.
Qed.

Lemma Bn_desc_2:
forall n:nat,
Bn n<=b0/2^n.
Proof.
induction n.
simpl; fourier.
simpl.
unfold Rdiv in IHn.
assert (1/2* Bn n<=b0*/(2*2^n)).
replace (b0*/(2*2^n)) with (1/2*(b0/2^n)).
apply Rmult_le_compat_l.
fourier.
assumption.
field.
apply Rgt_not_eq.
apply pow2_diff0.
apply Rle_trans with (1/2*Bn n).
apply An_bn_c12.
done.
Qed.

Lemma not_series:
forall n, dist_Rp (Xn n) (X0) <= 2*b0.
Proof.
have H1: forall n, dist_Rp (Xn (S n)) (Xn n) <= Bn n.
move => n.
rewrite vector.dist_sym.
unfold dist_Rp.
replace (dif_v (Xn  n) (Xn (S n))) with (mult_mv ( (f' (Xn n))^-1m) (f (Xn n))).
by  elim (about_seq n) => [_[_[_[H _]]]].
rewrite  //=.
 apply/vecP =>i; 
toR; rewrite vxE. 
rewrite !vxE. 
toR; ring.  
have H2: forall n, dist_Rp (Xn n) X0 <= 2*b0 - 2*b0/(2^n).
induction n.
rewrite //=  norm_eq.
replace (2*b0 - 2*b0/1) with 0; [|field; auto]; fourier. 
simpl (2^S n).
rewrite /dist_Rp.
replace (dif_v (Xn (S n)) X0) with (add_v (dif_v (Xn (S n)) (Xn n)) (dif_v (Xn n) X0)).
have H3: norm (dif_v (Xn (S n)) (Xn n)) <= b0/2^n.
apply Rle_trans with (Bn n).
apply (H1 n).
apply (Bn_desc_2).
apply Rle_trans with ( norm (dif_v (Xn (S n)) (Xn n)) + norm (dif_v (Xn n) X0)).
apply (norm_tri ).
replace (2*b0 - 2*b0/(2*2^n)) with (b0/2^n + (2*b0 - 2*b0/2^n)).
apply Rplus_le_compat; assumption.
unfold Rdiv.
field.
apply Rgt_not_eq.
apply pow2_diff0.
apply/vecP => i.
rewrite /add_v /dif_v; 
rewrite !vxE ; toR ; ring.
move => n.
apply Rle_trans  with  (2 * b0 - 2 * b0 / 2 ^ n).
exact (H2 n).
replace (2*b0-2*b0/(2^n)) with (2*b0 + (-(2*b0*/(2^n)))); [|field;auto].
pattern (2*b0) at 3; replace (2*b0) with (2*b0+0) ;[| ring].
apply Rplus_le_compat_l.
replace 0 with (-0);[| ring].
apply Ropp_le_contravar.
apply Rle_mult_inv_pos.
apply Rmult_le_pos; [fourier|apply b0_ge_0].
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.

Lemma seq_lim_in_dom:
forall xn (l:Rvec p) r,
0<=r->
conv xn l ->
(forall n, dist_Rp (xn n) (xn 0%nat) <= r)->
dist_Rp l (xn 0%nat) <= r.
Proof.
move =>xn l r Hr Hcv Hin.
have Hco: ~  r<dist_Rp l (xn 0%nat) .
unfold not => Hno.
have H1:  0 < dist_Rp l (xn 0%nat)-r.
fourier.
rewrite /conv /dist /= in Hcv.
have Hco1: exists N, forall n, (N<=n)%nat ->dist_Rp (xn n) l < dist_Rp l (xn 0%nat)-r .
apply (Hcv (dist_Rp l (xn 0%nat) - r) H1).
elim Hco1 => N Hlim.
have Hdist: dist_Rp (xn N) l < dist_Rp l (xn 0%nat) - r.
apply (Hlim N); apply/leP;omega.
have H3: r<r.
apply Rlt_le_trans with (dist_Rp (xn 0%nat) l - dist_Rp (xn N)l ).
rewrite vector.dist_sym; fourier.
apply Rle_trans with (dist_Rp (dif_v (xn 0%nat) l) (dif_v (xn N) l)).
apply (dist_min_tri  ).
replace (dist_Rp (dif_v (xn 0%nat) l) (dif_v (xn N) l)) with (dist_Rp (xn 0%nat) (xn N)).
rewrite vector.dist_sym; exact (Hin N).
rewrite /dist_Rp.
have H: (dif_v (xn 0%nat) (xn N))=(dif_v (dif_v (xn 0%nat) l) (dif_v (xn N) l)).
by apply/vecP => i; rewrite /dif_v !vxE; ring_simplify.
by rewrite H.
have H: ~(r<r); fourier.
elim (Rle_or_lt (dist_Rp l (xn 0%nat)) r) => H.
fourier.
contradiction.
Qed.

Lemma cont_eq_fun_Rp:
forall (f g: Rvec p -> Rvec p) a,
(forall x, f x = g x)->
cont  f a ->
cont g a.
Proof.
by move=> h g a Hfg Hf;
rewrite /cont in Hf |-* =>eps He; elim (Hf eps He)=> alp [Ha1 Ha2]; 
exists alp; split; [done|move=>x Hx; rewrite -!Hfg; apply Ha2].
Qed.

(*Notation "\vec_ ( i ) E" :=
  (@vec_of_fun p (fun i  => E)) (at level 25).
*)

Lemma unic_aux:
forall xs2:Rvec p,
  norm (dif_v xs2 X0) <= 2*b0-> 
  f xs2 = vect0 ->
 forall n, norm (dif_v xs2  (Xn n)) <= 2*Bn n.
Proof.
intros xs2 Hdisc2 Hrad.
induction n.
simpl.
assumption.
assert (Hf'_not_zero : \det (f' (Xn n)) <> 0).
destruct (about_seq n) as [_ [Hf' _]]; intros; assumption.
assert (Hr_seq2: Xn (S n) -^ xs2 =
( mult_mv ((f' (Xn n))^-1m) 
((f xs2 -^ f (Xn n)) -^ mult_mv (f' (Xn n)) ((xs2 -^ Xn n))))). 
rewrite Hrad.
simpl.
have H1: (vect0  -^ f (Xn n)) =  -^ f (Xn n).
by eq_vec.
rewrite H1; clear H1.
have H1: mult_mv (p:=p) ((f' (Xn n))^-1m)
  ((-^ f (Xn n)) -^ mult_mv (p:=p) (f' (Xn n)) (xs2 -^ Xn n)) = 
((-^ mult_mv ((f' (Xn n))^-1m)  (f (Xn n)))) -^ mult_mv ((f' (Xn n))^-1m)
 (mult_mv (p:=p) (f' (Xn n)) (xs2 -^ Xn n)).
by rewrite -opp_distr mult_mv_assoc 
 -distr_opp_vm opp_distr mult_mv_assoc.
rewrite H1 mult_mv_assoc.
replace 
( (@mulmx real_ringType  p p p
           (@invmx real_comUnitRingType p (f' (Xn n))) (f' (Xn n))))

with (((f' (Xn n))^-1 * f' (Xn n)))%Ri.


(*replace ((f' (Xn n))^-1m *m f' (Xn n)) with (((f' (Xn n))^-1 * f' (Xn n)))%Ri.*)
rewrite  mulVr.
rewrite unitM_el_neu.
by eq_vec.
by move/eqP: Hf'_not_zero.
by toR.
case X: (\det (f' (Xn n)) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := Hf'_not_zero X.
done.
rewrite -/(dist_Rp xs2 (Xn (S n))) vector.dist_sym /dist_Rp Hr_seq2.
apply Rle_trans with (norm_m ((f' (Xn n))^-1m) * 
  norm ((f xs2 -^ f (Xn n)) -^ mult_mv (p:=p) (f' (Xn n)) (xs2 -^ Xn n))).
apply ineq_norm_mv.
simpl.
rewrite -/(INR p).
replace (2*(INR p *An n* Bn n*Bn n*c)) with ( An n*(2*INR p *Bn n*Bn n*c));
last by  ring.
apply Rmult_le_compat.
apply normM_pos.
apply norm_pos.
by destruct (about_seq n) as  [_ [_[Hf' _]]].
apply Rle_trans with (1/2*INR p*Rsqr(norm (xs2 -^ Xn n))*c).
have H1: forall r, ((f xs2 -^ f (Xn n)) -^ r) = 
  (f xs2 -^ (f (Xn n) +^ r)). move=> r; by eq_vec .
rewrite H1.
apply ineq12.
apply Rle_trans with (1 / 2 * INR p * Rsqr (2*Bn n) * c ).
apply Rmult_le_compat_r; first by apply c_ge_0.
apply Rmult_le_compat_l.
apply Rmult_le_pos; first by fourier.
apply pos_INR.
apply Rsqr_incr_1; [done| apply norm_pos|apply Rmult_le_pos;[fourier|apply Bn_gt_0]].
rewrite /Rsqr; right; field.
Qed.
 

Lemma Bn_monot:
forall n, Bn (S n)<=Bn n.
Proof.
induction n.
simpl.
rewrite -/(INR p).
replace (INR p*a0*b0*b0*c) with (b0*(INR p*a0*b0*c)) ; last by ring.
pattern b0 at 3; replace b0 with (b0*1) ; last by ring.
apply Rmult_le_compat_l ; [apply b0_ge_0|].
apply Rle_trans with (2*INR p*a0*b0*c); last by done.
pattern (INR p*a0*b0*c) at 1; replace (INR p*a0*b0*c) with (1*INR p*a0*b0*c)
  ; last by ring.
repeat rewrite Rmult_assoc.
apply Rmult_le_compat_r; last by fourier.
repeat apply Rmult_le_pos;
   [apply pos_INR| apply a0_ge_0| apply b0_ge_0|apply c_ge_0].
replace (Bn (S (S n))) with ((Bn (S n)*(INR p*An (S n)*Bn (S n)*c))); [|simpl;ring].
pattern (Bn (S n)) at 3; replace (Bn (S n)) with (Bn (S n)*1) ; last by ring.
apply Rmult_le_compat_l.
apply Bn_gt_0.
apply Rle_trans with (2*(INR p*An (S n)*Bn (S n)*c)).
pattern (INR p*An (S n)*Bn (S n)*c) at 1; replace (INR p*An (S n)*Bn (S n)*c) with 
  (1*(INR p*An (S n)*Bn (S n)*c)); last by ring.
apply Rmult_le_compat_r; last by fourier.
apply Rmult_le_pos; last by apply c_ge_0 .
apply Rmult_le_pos; last by apply Bn_gt_0 .
apply Rmult_le_pos; [ apply pos_INR| apply An_gt_0].
replace (2*(INR p*An (S n)*Bn (S n)*c)) with (2*INR p*An (S n)*Bn (S n)*c) ; last by ring.
apply An_bn_c1.
Qed.

Lemma Bn_monot2:
forall (n p:nat), Bn (n+(S p))%nat <= Bn n.
Proof.
intros n p0.
induction p0.
replace (n+1)%nat with (S n) .
apply Bn_monot.
by nat_norm.
replace (n+S (S p0) )%coq_nat with (S (n+(S p0))%coq_nat) ; last by omega.
apply Rle_trans with (Bn (n+(S p0)))%coq_nat; last by done.
nat_norm; apply Bn_monot.
Qed.

Lemma Bn_monot3:
forall (n p:nat), ( (S p)<=n)%coq_nat-> Bn n<=Bn  p.
Proof.
intros n p0 Hlt.
replace n with (p0+S(n- (S p0))%coq_nat)%coq_nat.
apply Bn_monot2.
omega.
Qed.


Lemma Bn_conv_0:
Un_cv Bn 0.
Proof.
red.
intros eps0 He0.
assert (Hb01:  exists n:nat, (b0/(2^n) < eps0)).
apply archim3_lt;[fourier|apply b0_ge_0].
elim Hb01; intros N Hlt.
exists (S (S N)).
intros n Hineq.
apply Rle_lt_trans with (b0/2^N); last by done.
unfold R_dist.
replace (Bn n - 0) with (Bn n) ; last by ring.
rewrite Rabs_pos_eq; [|apply Bn_gt_0].
apply Rle_trans with (Bn (S N)).
apply (Bn_monot3 ).
omega.
apply Rle_trans with (Bn N).
apply Bn_monot.
apply Bn_desc_2.
Qed.

Theorem kantoro_unic:
forall xs2:Rvec p,
 norm (dif_v xs2 X0) <= 2*b0 ->
  f xs2 = vect0 ->
  conv Xn xs2.
Proof.
intros xs2 Hdisc2 Hrad.
assert (Hseq:= @unic_aux xs2 Hdisc2 Hrad).
red.
intros delta Hdpos.
assert (HBn_cv: Un_cv Bn 0); [apply Bn_conv_0|].
unfold Un_cv in HBn_cv.
assert (Hd2: delta/2>0).
fourier.
elim (HBn_cv (delta/2) Hd2); intros N Hbcv.
exists N.
intros n .
move/leP => Hineq.
rewrite /dist /= vector.dist_sym.
apply Rle_lt_trans with (2*Bn n).
apply (Hseq n).
assert (Hbn_d: R_dist (Bn n) 0<delta/2).
apply (Hbcv n Hineq).
unfold R_dist in Hbn_d.
replace (Bn n-0) with (Bn n) in Hbn_d ; last by ring.
rewrite Rabs_right in Hbn_d.
fourier.
apply Rle_ge; apply Bn_gt_0.
Qed.


Theorem kantoroRp_exists:
exists xs:Rvec p,
conv Xn xs /\ norm (dif_v xs X0) <= 2*b0 /\ f xs = vect0.
Proof.
have Hcc: isCauchy Xn.
have Hcc1: forall n,
  (n<>0%nat-> forall v, (dist_Rp (Xn n) v <= (En n))->
    (dist_Rp (Xn (Peano.pred n)) v <= (En (Peano.pred n)))).
move =>n.
by elim (about_seq n) => H.
rewrite Cauchy_crit_Rp2.
move => eps0 He.
have Hincl:forall n p, (dist_Rp (Xn n) (Xn (n+p)%nat) <= En n ).
apply disc_incl_pct.
intros n.
apply (Hcc1 (S n)).
omega.
assert (Harch: exists n:nat, (keps/(2^n) < eps0)).
apply archim2_lt; fourier.
elim Harch; intros N Heps.
exists N; intros n p0 H1.
assert (Hnp:dist_Rp (Xn n) (Xn (n+p0)%nat) <= En n).
apply (Hincl n p0).
rewrite transf_en in Hnp.
assert (keps/2^n<=keps/2^N).
apply (pow2_monot3 ); assumption.
apply Rle_lt_trans with (keps/2^N).
by apply Rle_trans with (keps/2^n).
done.
elim (Rp_complete Hcc) => xs Hconv.
exists xs.
split.
done.
split.
fold (dist_Rp xs X0).
have H1:= not_series.
replace X0 with (Xn 0%nat).
apply seq_lim_in_dom.
apply Rmult_le_pos; [fourier|apply b0_ge_0].
by done.
exact H1.
by rewrite //=.
assert (Hxn: forall n, 
f (Xn n) = mult_mv (f' (Xn n))  ((Xn n) -^ (Xn (S n)))).
have H1: forall n,  mult_mv ((f' (Xn n))^-1m) (f (Xn n)) = ((Xn n) -^ (Xn (S n))).
intros n.
simpl.
rewrite /dif_v ; apply/vecP => i; rewrite !vxE; 
toR; ring.
move=> n; rewrite -(H1 n) mult_mv_assoc.
destruct (about_seq n) as [_ [Hf' _]].
have Hf'' : \det (f' (Xn n)) !=0.
case X: (\det (f' (Xn n)) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := Hf' X.
done.
replace ((f' (Xn n) *m (f' (Xn n))^-1m) ) with ((f' (Xn n) * (f' (Xn n))^-1))%Ri;
  last by toR.
rewrite mulrV; last by done.
by rewrite unitM_el_neu.
have Hcv: conv (fun i:nat=> mult_mv (f' (Xn i)) ((Xn i) -^ (Xn (S i)))) (vect0).
apply lim_prod_minus with xs;[|done].

move=> a i.
apply  rel_cont_R_Rp => j.
apply  cont_eq_fun with ((fun x  => @der_p p  f j x (pr x j) i)).
move=> x.
by rewrite Hjac /Jac !vxE !mxE.
by apply rel_cont_Rp.
have Hcv1: conv (fun i:nat =>f (Xn i)) vect0 .
intros eps0 He0.
elim (Hcv eps0 He0); intros N Habs.
exists N.
intros n Hneq.
rewrite (Hxn n).
exact (Habs n Hneq).
have Hcv2: conv (fun i:nat => f (Xn i)) (f xs).
apply Heine1_Rp ;  done.
by apply: (@seq_lim_Rp_uniq  _ (fun i : nat => f (Xn i))).
Qed.

Definition mu n := 2*INR p*An n*Bn n*c.

Open Scope R_scope.

Lemma rel_mu:
forall n, mu (S n) = pow (mu n) 2.
Proof.
induction n; unfold mu;  simpl; ring.
Qed.

Fixpoint Npow (r:nat) (n:nat) {struct n} : nat :=
  match n with
    | O => 1%nat
    | S n =>  (r  * (Npow r n))%nat
  end.




Lemma rel_mu0_aux:
forall p, mu p =  (mu 0) ^ (Npow 2 p).
Proof.
move=> p0; induction p0; [unfold mu; simpl; ring|].
rewrite rel_mu; rewrite IHp0 ;
rewrite <- pow_mult.
assert (Heq: (Npow 2 p0 * 2)%nat = Npow 2 (S p0)).
by simpl; rewrite mulnC.
rewrite multE Heq; trivial.
Qed.

Lemma eq_pow:
forall r n, INR (Npow r n)  = (INR r) ^ n.
Proof.
intros r n.
induction n; simpl; [trivial
   |rewrite mult_INR; rewrite IHn; trivial].
Qed.

Lemma Npow_mon1:
forall (x n:nat) , (1 <= x)%nat -> (Npow x  n <= Npow x (S n))%nat.
Proof.
intros x n Hx; induction n.
by rewrite //= muln1.
simpl in IHn |-*.
rewrite leq_mul2l; apply/orP; by right.
Qed.



Lemma Npow_monot:
forall x m n, (1 <= x)%nat -> 
(Npow x (m) <= Npow x (m+n)%nat)%nat.
Proof.
intros x m n Hx .
induction n.
by replace (m + 0)%nat with m.
apply leq_trans with (Npow x (m + n)%nat).
trivial.
replace (m+(S n))%nat 
  with (S (m+n)%nat) .
apply Npow_mon1; trivial.
by nat_norm.
Qed.


Infix Local "^R" := Rpower (at level 30, right associativity) : R_scope.

Lemma rel_bn1:
forall n, Bn (S n) = /2*(mu 0)^(Npow 2 n)*Bn n.
Proof.
intro n; simpl; rewrite -/(INR p);
rewrite <- rel_mu0_aux; unfold mu; field.
Qed.

Lemma rel_bn : 
forall n, Bn n = /2^n * (mu 0)^((Npow 2 n)-1)%coq_nat * b0 .
Proof.
induction n; [simpl;field|].
rewrite rel_bn1; rewrite IHn; simpl.
field_simplify.
rewrite <- pow_add; rewrite Rmult_comm.
assert (Heq: (Npow 2 n + (Npow 2 n - 1)%coq_nat)%coq_nat  =
             ((Npow 2 n + (Npow 2 n + O%coq_nat)%coq_nat)%coq_nat -1)%coq_nat).
omega.
rewrite Heq.
 trivial.
apply Rgt_not_eq; apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.

Lemma pow_le1:
forall x n, 0 <=x <=1 -> 0 <= x^n <= 1.
Proof.
intros x n [H0 H1].
induction n; simpl; [split; [fourier|right; trivial]| elim IHn; intros IH1 IH2].
split. 
replace 0 with (0*x^n); last by ring.
auto with real.
replace 1 with (1*1) ; last by ring.
apply Rmult_le_compat; assumption.
Qed.

Lemma pow_le1_le:
forall x (m n: nat), 0 <= x <= 1 -> (0 <= n <= m)%nat -> x^m <=x^n.
Proof.
intros x m n [H0 H1] Hmn.
replace m with (m - n + n)%nat .
  rewrite pow_add.
  pattern (x ^ n) at 2 in |- *; replace (x ^ n) with (1 * x ^ n);
    auto with real.
  apply Rminus_le.
  repeat rewrite (fun y:R => Rmult_comm y (x ^ n));
    rewrite <- Rmult_minus_distr_l.
  replace 0 with (x ^ n * 0); auto with real.
  apply Rmult_le_compat_l; auto with real.
   apply pow_le; auto with real.
apply Rle_minus.
exact (proj2 (pow_le1 (m-n) (conj H0 H1))).
by rewrite subnK.
Qed.

Lemma Npow_gt1:
forall x n, (0 <x)%nat -> (1 <= Npow x n)%nat .
Proof.
intros x n Hx; induction n; simpl; apply/leP; [omega|].
apply le_trans with (Npow x n); [by move/leP: IHn => IHn|].
elim (mult_O_le (Npow x n) x) ; intros H0; [|trivial] .
rewrite lt0n in Hx.
move/eqP: Hx => Hx.
by elim Hx.  
Qed.

(*
Lemma minus_le_compat:
forall m n p, (m <= n)%nat -> (m-p <= n - p)%nat.
Proof.
intros m n k H.
move/leP: H => H.
apply/leP.
elim H.
omega.
intros m0 H1 H2.
apply le_trans with (m0 - k)%nat.
done.

apply/leP; apply leqnSn.


Qed.
*)


Lemma Hdif_pos: forall n k, norm (Xn (k + n)%nat -^ Xn k)<=2*Bn k - (2*Bn k)/2^n.
Proof.
intros n k; induction n.
simpl; rewrite addn0.
rewrite  /Rdiv Rinv_1; ring_simplify.
replace (Xn k -^ Xn k) with (@vect0 p);  last by eq_vec.
have H0:= refl_equal (@vect0 p).
move/norm0 : H0 => H0.
by right.
simpl.
replace (Xn (k + S n) -^ Xn k) with 
  (Xn (k + S n) -^ Xn (k+n) +^ (Xn (k + n) -^ Xn k));
last by eq_vec.
assert (H0 : norm (Xn (k + (S n)) -^ Xn ( k+ n)) <= Bn k / 2^n).
apply Rle_trans with (Bn (k+n)).
replace (k + S n)%nat with (S (k + n))%nat ; last by nat_norm. 
fold (dist_Rp(Xn (k + n).+1) (Xn (k + n)) ).
rewrite vector.dist_sym.
destruct (about_seq (k+n)) as [_[_[_ [Hineq _]]]].
rewrite /dist_Rp.
replace (Xn  (k + n) -^ Xn (S (k + n)%nat)) with
(mult_mv ((f' (Xn (k + n)))^-1m) (f (Xn (k + n)))).
destruct (about_seq (k+n)) as [_[Hf'[_]]]; intros; assumption.
simpl; rewrite /dif_v; apply/vecP => i; rewrite !vxE /=; ring.
repeat rewrite rel_bn.
field_simplify; [|split; apply Rgt_not_eq; apply pow2_diff0|
   apply Rgt_not_eq; apply pow2_diff0].
rewrite Rmult_comm; rewrite <-pow_add.
unfold Rdiv; repeat rewrite Rmult_assoc.
apply Rmult_le_compat_l; [apply b0_ge_0|].
apply Rmult_le_compat_r.
left; apply Rinv_0_lt_compat; apply pow2_diff0.
apply pow_le1_le.
split; rewrite /mu /=.
 repeat apply Rmult_le_pos. 
 fourier.
rewrite -/(INR p); apply pos_INR.
apply a0_ge_0.
apply b0_ge_0.
apply c_ge_0.
done.
apply/leP.
assert (H02: (0<2)%nat).
apply/ltP; omega.
assert (H1:= Npow_gt1 k H02).
move/ltP: H1=> H1.
apply/leP.
rewrite (minusE (Npow 2 k) (S 0)) (minusE (Npow 2 (k + n)) (S 0)).
apply leq_sub2r .
by apply Npow_monot.
apply Rle_trans with (norm (Xn (k + S n) -^ Xn (k + n)) 
  +  norm(Xn (k + n) -^ Xn k)).
apply: norm_tri.
replace (2*Bn k - 2*Bn k/(2*2^n)) with (Bn k/2^n + (2*Bn k - 2*Bn k/2^n)).
apply Rplus_le_compat; assumption.
unfold Rdiv; field; apply Rgt_not_eq; apply pow2_diff0.
Qed.
Lemma lim_conv_up:
forall (un:nat->R) (a l:R),
(forall n, un n <=a)->
Un_cv un l ->
l<=a.
Proof.
intros un t l H H0.
cut (~(l>t)).
intros H1.
apply Rnot_gt_le; auto.
unfold not; intros H1.
unfold Un_cv in H0.
assert (l-t>0) by fourier.
assert (Hcon: exists N, forall n, (n>=N)%coq_nat ->R_dist (un n) l <l-t) by
apply (H0 (l-t) H2).
elim Hcon; intros N Hlim.
assert (Hdist:R_dist (un N) l<l-t).
apply (Hlim N); omega.
unfold R_dist in Hdist.
assert (Ha1: un N <= t) by apply (H N).
rewrite Rabs_minus_sym in Hdist.
rewrite Rabs_right in Hdist; [|fourier].
assert (un N>t); fourier.
Qed.

Lemma norm_dif_tri:
forall a b :Rvec p,
(norm a - norm b) <= norm (a -^ b).
Proof.
move=> a b.
apply (Rplus_le_reg_l (norm b)); ring_simplify.
by pattern a at 1; replace a with (b +^ (a -^ b));
[ apply norm_tri |eq_vec].
Qed.


Lemma norm_dif_tri2:
forall a b :Rvec p,
Rabs (norm a - norm b) <= norm (a -^ b).
Proof.
move=> a b.
rewrite /Rabs; case (Rcase_abs  (norm a - norm b) ) => *;
[rewrite Ropp_minus_distr -/(dist_Rp a b) vector.dist_sym /dist_Rp|];
apply norm_dif_tri. 
Qed.


Lemma conv_speed:
forall k (xs:Rvec p),
 norm (dif_v xs X0) <= 2*b0 -> f xs = vect0  -> 
norm (xs -^ Xn k) <= (/ 2^R (INR k - 1)) * mu 0 ^ (Npow 2 k - 1) * b0 .
Proof.
move=> k xs Hdom Hroot.
apply  lim_conv_up with (fun n => norm (Xn (k+n) -^ Xn k)).
intro n; apply Rle_trans with (2*Bn k - (2*Bn k)/2^n) ; [apply Hdif_pos|].
rewrite rel_bn.
rewrite (minusE (Npow 2 k)  (S 0)).
replace (/ 2 ^R (INR k - 1) * mu 0 ^ (Npow 2 k - 1) * b0) with 
(/ 2 ^R (INR k - 1) * mu 0 ^ (Npow 2 k - 1) * b0+(-0)); last by ring.
replace (2 * (/ 2 ^ k * mu 0 ^ (Npow 2 k - 1) * b0)) with
 (/ 2 ^R (INR k - 1) * mu 0 ^ (Npow 2 k - 1) * b0).
unfold Rminus.
 apply Rplus_le_compat_l.
apply Ropp_le_contravar.
unfold Rdiv.
repeat  apply Rmult_le_pos.
left; apply Rinv_0_lt_compat.
unfold Rpower; apply exp_pos.
assert (H: 0 <= mu 0 <= 1).
    unfold mu; simpl; rewrite -/(INR p); split.
repeat  apply Rmult_le_pos;
[fourier| apply pos_INR|apply a0_ge_0|apply b0_ge_0|apply c_ge_0].
done.
apply (proj1 (pow_le1 (Npow 2 k - 1) H)).
apply b0_ge_0.
left; apply Rinv_0_lt_compat; apply pow2_diff0.
repeat rewrite <-Rmult_assoc.
replace (2 * / 2 ^ k ) with (/2 ^R ( INR k - 1)); [trivial|].
unfold Rminus.
rewrite Rpower_plus.
rewrite Rpower_Ropp.
rewrite Rpower_1; [|fourier].
rewrite Rpower_pow; [|fourier].
field; apply Rgt_not_eq; apply pow2_diff0.
intros e He.
have Hcv:= kantoro_unic Hdom Hroot.
elim (Hcv e He) ; intros N HN.
exists N; intros n Hn.
apply Rle_lt_trans with (norm ( (Xn (k + n) -^ Xn k) -^  (xs -^ Xn k))).
apply: norm_dif_tri2.
replace ((Xn (k + n) -^ Xn k) -^ (xs -^ Xn k)) with  ((Xn (k+n)) -^ xs).
rewrite /dist /= /dist_Rp in HN.
apply HN; rewrite -plusE; apply/leP; omega.
by eq_vec.
Qed.



End Kantorovitch.

                                                           

