Add LoadPath "./multi_thesis".

Require Import ssreflect ssrbool ssrnat fintype ssrfun bigops matrix.
Require Import Rstruct metric min_max vector.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.
Require Import ClassicalChoice.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section sequence.
Variable p: nat.
 Notation Local " 'Rvec' p " := (vec R p)(at level 20): type_scope.

Notation "v1 +^ v2" := (add_v v1 v2) (at level 50).
Notation "a *^ v" := (mult_sv a v) (at level 40).
Notation "v1 -^ v2" := (dif_v v1 v2) (at level 50).
Notation "-^ v" := (opp_v v) (at level 35).

(*
Definition conv_Rp (un:nat->Rvec p) (l:(Rvec p)) : Prop :=
    forall eps:R,
      R0 < eps->
      exists N : nat, (forall n:nat, (N <= n)%coq_nat -> dist_Rp (un n) l < eps).

Definition Cauchy_crit_Rp (vn:nat->(Rvec p)): Prop :=
    forall eps:R,
      R0 < eps ->
      exists N : nat,
        (forall n m:nat,
          (N <= n)%coq_nat -> (N <= m)%coq_nat -> dist_Rp (vn n) (vn m) < eps).
*)

Lemma cond_eq_Rp :
  forall (x y: Rvec p), (forall eps:R, R0 < eps -> dist_Rp x  y < eps) -> x = y.
Proof.
move => x y H.
have Hra: ~ R0 < dist_Rp x y.
rewrite /not => Hdist.
have Hco: dist_Rp x  y < (dist_Rp x y)*/2.
by apply H; fourier.
have Hco2: ~ dist_Rp x y < dist_Rp x y */2.
by fourier.
contradiction.
rewrite -vector.dist_refl.
by elim (vector.dist_pos x y).
Qed.

Require Import Max.

Lemma seq_lim_Rp_uniq:
forall (vn: nat -> Rvec p) l l', conv vn l -> conv vn l' -> l = l'.
Proof.
move => vn l l' Hcvl Hcvl'.
rewrite /conv in Hcvl Hcvl'.
apply cond_eq_Rp => eps Hepos.
elim (Hcvl (eps*/2) (eps2_Rgt_R0 eps Hepos)) => N1 Hcv1.
elim (Hcvl' (eps*/2) (eps2_Rgt_R0 eps Hepos)) => N2 Hcv2.
rewrite -(eps2 eps) /dist_Rp.
have Heq: forall n:nat, dif_v  l l' = add_v (dif_v  l (vn n)) (dif_v (vn n) l').
by move => n; eq_vec.
(*  rewrite - ext_eq ;rewrite /dif_v /add_v => i; rewrite !vxK //=; field.*)
rewrite (Heq (max N1 N2)); apply Rle_lt_trans with
 (norm (dif_v l (vn (max N1 N2)))+norm (dif_v (vn (max N1 N2)) l')).
apply norm_tri.
apply Rplus_lt_compat.
fold (dist_Rp l (vn (max N1 N2)));rewrite vector.dist_sym; apply (Hcv1 (max N1 N2)).
apply/leP; apply le_max_l.
fold (dist_Rp (vn (max N1 N2)) l');apply (Hcv2 (max N1 N2));  apply/leP; apply le_max_r.
Qed.

Lemma CV_Cauchy_Rp :
  forall vn:nat -> Rvec p, sigT (fun l => conv vn l) -> isCauchy vn.
Proof.
move => vn Hconv.
elim Hconv => l Hcv.
rewrite /isCauchy =>eps Hepos.
rewrite /conv in Hcv.
elim (Hcv (eps*/2) (eps2_Rgt_R0 eps Hepos)) => N Hdis.
exists N.
move=> n m Hn Hm.
apply Rle_lt_trans with (dist_Rp (vn n) l + dist_Rp l (vn m)).
by apply: vector.dist_tri.
rewrite -(eps2 eps);apply Rplus_lt_compat; [exact (Hdis n Hn)|rewrite vector.dist_sym;
exact (Hdis m Hm)].
Qed.

Lemma comp_cauchy:
forall (v:nat->(Rvec p)), 
isCauchy v ->
forall k:'I_p, Cauchy_crit (fun n => v n k).
Proof.
move => v  HcauRp k.
rewrite /Cauchy_crit => eps Hepos.
rewrite /isCauchy in HcauRp.
elim (HcauRp eps Hepos)=> N Hcau.
exists N.
move=> n m Hn Hm.
have Hn': (N<=n)%coq_nat.
 by omega.
have Hm': (N<=m)%coq_nat.
  by omega.
rewrite /R_dist.
move/leP: Hn' => Hn'.
move/leP: Hm' => Hm'.
have HRp_dis:= (Hcau n m Hn' Hm').
rewrite /dist_Rp in HRp_dis.
apply Rle_lt_trans with (norm (dif_v (v n) (v m))).
have H: (v n k - v m k = (dif_v (v n) (v m))  k)%R.
by rewrite /dif_v vxE.
rewrite H. 
 apply norm_max.
done.
Qed.
 
Lemma lim_comp:
forall (v:nat->(Rvec p)),
isCauchy v->
forall k:'I_p, { l:R | Un_cv (fun n => v n k) l}.
Proof.
move => v Hcau k.
apply R_complete.
by apply comp_cauchy.
Qed.

Definition lim_Rp_seq vn (Hcau: isCauchy vn) :=
(\vec_(k) projT1 (lim_comp Hcau k)).



Lemma equiv_conv:
forall (vn:nat -> Rvec p) l,
conv vn l <-> forall k, Un_cv (fun n => vn n k) (l k). 
Proof.
move => vn l.
split =>Hcv.
move =>k.
rewrite /Un_cv.
move =>eps Hepos.
rewrite /conv in Hcv.
elim (Hcv eps Hepos)=> N Hcv1.
exists N.
move => n Hn.
move/leP: Hn => Hn.
(*have Hn': (N<=n)%coq_nat.
by omega.
*)
rewrite /R_dist.
have Hdis:=(Hcv1 n Hn).
rewrite /dist_Rp in Hdis.
apply Rle_lt_trans with (norm (vn n -^ l)).
have H: (vn n k - l k = (vn n -^ l)  k)%R.
by rewrite /dif_v vxE.
rewrite H. 
by apply norm_max.
done.
rewrite /conv =>eps Hepos.
rewrite /Un_cv in Hcv.
have Hcv': forall (eps:R), eps >R0 -> forall (k :'I_p), 
  exists N:nat, forall n:nat, (n>=N)%coq_nat -> R_dist (vn n k) (l k) <eps.
move => eps0 He0 k; exact (Hcv k eps0 He0).
have Hepos': eps > R0 by fourier.
have Hcv_c:= (Hcv' eps Hepos').
elim (choice (fun k N =>forall n, (n>=N)%coq_nat->R_dist (vn n k) (l k)<eps) Hcv_c)
=> f Hrel.
exists (\max_(i < p) f i).
move => n Hmax.
rewrite /dist /= /dist_Rp .
have H: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n0; by right.
elim: H => H.
rewrite /norm.
rewrite !big_ip_zero; try done . 
elim (norm_rew H (dif_v (vn n) l)) => k Hnorm.
rewrite Hnorm /dif_v vxE.
have Hfk: (f k <= n)%coq_nat.
 apply le_trans with ( \max_(i < p) f i).
apply/leP.
 apply: leq_bigmax.
by apply/leP.
have Hfk': (n >= f k)%coq_nat by omega.
exact (Hrel k n Hfk').
Qed.
 
 
Lemma Rp_complete:
forall (vn: nat -> Rvec p), isCauchy vn -> exists l, conv vn l.
Proof.
move => vn Hcau.
exists (lim_Rp_seq Hcau).
rewrite equiv_conv.
move => k.
rewrite /lim_Rp_seq vxE.
case (lim_comp Hcau k).
rewrite //=.
Qed.

Lemma dist0_eq:
forall (x:Rvec p),
dist_Rp x x = R0.
Proof.
move => x; by  rewrite vector.dist_refl.
Qed.

End sequence.

