Add LoadPath "./multi_thesis".
 
Require Import ssreflect ssrbool ssrnat eqtype seq
  choice fintype ssrfun bigops finfun tuple matrix.
Require Import Rstruct min_max topology.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
(*Require Import Classical.*)

(* this implementation was originally designed for real vectors and the 
development of multivariate analysis;
maybe i should make the vectors more general, similar to the matrix 
library: defined on a generic type, eqType , ring etc. and with the 
approproate properties in each case *)

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Reserved Notation "\vec_ ( i ) E"   (at level 36, E at level 36, i at level 50,
   format "\vec_ ( i  )  E").

Reserved Notation "\vec_ ( i < p ) E"
  (at level 36, E at level 36, i, p at level 50,
  format "\vec_ ( i  <  p )  E").

Delimit Scope vector_scope with V.

Open Scope vector_scope.

(*Open Local Scope ring_scope.
Open Scope matrix_scope.
*)

Section VectorDef.

Variable R : Type.
Variable p: nat.
(*Notation p := p'.+1.*)
(*
Lemma Hp: (0<p)%Dn.
by done. Qed.
*)

CoInductive vec : predArgType := Vector of {ffun 'I_p -> R}.

Definition vval A := let: Vector g := A in g.

Lemma vvalK : cancel vval Vector.  
by rewrite /cancel; case => /=. Qed.

Definition vec_of_fun F := locked Vector [ffun i => F i].

Definition fun_of_vec A (i : 'I_p) := vval A i.

Coercion fun_of_vec : vec >-> Funclass.

Lemma vxE : forall F, vec_of_fun F =1 F.
Proof. by unlock vec_of_fun fun_of_vec => F i ; rewrite /= ffunE . Qed.

Lemma vecP : forall A B : vec, A =1 B <-> A = B.
Proof.
unlock fun_of_vec => [] [A] [B]; split=> [/= eqAB | -> //].
congr Vector; apply/ffunP=> [] [i j]; exact: eqAB.
Qed.


End VectorDef.


Definition vec_eqMixin (eT : eqType) p:=
  CanEqMixin (@vvalK eT p).
Canonical Structure vec_eqType (eT: eqType) p:=
  Eval hnf in EqType (vec eT p) (vec_eqMixin eT p).


Definition vec_choiceMixin (R : choiceType) p :=
  CanChoiceMixin (@vvalK R p).
Canonical Structure vec_choiceType (R: choiceType) p :=
  Eval hnf in ChoiceType (vec R p) (vec_choiceMixin R p).

(*
Section Vector.

(*definition and properties for vectors of length p*)
Variable (p: pos_nat) (T: Type).

CoInductive Vec : predArgType := 
 Vect of {ffun 'I_p -> T}.
  
Definition vval v := match v with Vect g => g end.

Lemma can_vval : cancel vval Vect.
Proof. by rewrite /cancel; case => /=. Qed.

Lemma vval_inj : injective vval. 
Proof. exact: can_inj can_vval. Qed.


Definition vec_of_fun F :=
  locked Vect [ffun i => F i].

Definition fun_of_vec := locked (fun A i  => vval A i).

Coercion fun_of_vec : Vec >-> Funclass.

Lemma vxK : forall F, vec_of_fun F =1 F.
Proof. by unlock vec_of_fun fun_of_vec => F i ; rewrite /= ffunE . Qed.

End Vector.


Section EqVector.

Variable (p: pos_nat) (eT: eqType).

Definition vec_eqMixin := 

Canonical Structure vec_eqType := CanEqType (@can_vval p eT).

Lemma ext_eq : forall A B : Vec p eT, A =1 B <-> A = B.
Proof.
unlock fun_of_vec => [] [A] [B]; split=> [/= eqAB | -> //].
congr Vect; apply/ffunP=> [] [i j]; exact: eqAB.
Qed.

Lemma ext_eq1: forall u v : Vec p eT, u =1 v -> u = v.
Proof.
by move => u v Huv; rewrite -ext_eq.
Qed.

End EqVector.

*)


Notation "\vec_ ( i  < p ) E" :=
  (@vec_of_fun _ p (fun i => E)) (only parsing) : vector_scope.  

Notation "\vec_ ( i ) E" := (\vec_(i < _) E) : vector_scope.

Section Rvector.
(*operations on vectors*)

Variable p: nat.
(*Notation p := p'.+1.*)
Notation Local " 'Rvec' p " := (vec R p)(at level 20): type_scope.
Open Scope R_scope.
(*operations on vectors*)


Definition add_v (v1 v2: Rvec p):= \vec_( i < p) (v1 i + v2 i).
Definition mult_sv (a:R) (v:Rvec p):=  \vec_(i < p) (a* v i).
Definition dif_v (v1 v2: Rvec p):= \vec_(i) (v1 i - v2 i).
Definition opp_v (v:Rvec p):=\vec_(i) ( -v i).

(*canonical base*)
Definition vect0:= (\vec_(i < p) R0):Rvec p.
Definition base_v k:=(\vec_(i < p)(if k==i then R1 else R0)):Rvec p.


(*Other definitions*)
Definition Vabs v := \vec_(i < p)  Rabs (v i).
Definition Vle (v1 v2: vec R p) := forall i, v1 i <= v2 i.
Definition Vlt (v1 v2: vec R p) := forall i, v1 i < v2 i.

End Rvector.

Implicit Arguments dif_v [p].
Implicit Arguments add_v [p].
Implicit Arguments opp_v [p].
Implicit Arguments  mult_sv [p].
Implicit Arguments base_v [p].
Implicit Arguments vect0[p].
Implicit Arguments Vabs[p].
Implicit Arguments Vle[p].
Implicit Arguments Vlt[p].

Infix "<=v:" := Vle (at level 70): vector_scope.



Ltac eq_vec:=
match goal with 
| |- ( _ = _ ) => rewrite /add_v /mult_sv /dif_v /opp_v /vect0 /base_v;
                       apply/vecP => fresh; rewrite !vxE; ring_simplify
end.
(* 
| _ =>  rewrite /(@Ring.opp ring_add_grp ) /=
                        /(@Ring.add ring_add_grp ) /= 
                        /(@Ring.mul ring_base ) /=
                       /0 /= /1 /=
end.
*)

Section vectors.

(*Hypothesis Hp: (O<p)%nat.*)

Variable p: nat.
(*Notation p := p'.+1.*)




Notation "v1 +^ v2" := (@add_v  p v1 v2) (at level 50).
Notation "a *^ v" := (@mult_sv p a v) (at level 40).
Notation "v1 -^ v2" := (@dif_v p v1 v2) (at level 50).
Notation "-^ v" := (@opp_v p v) (at level 35).

Lemma mult_add: 
forall a v1 v2 ,
a *^ (v1 +^ v2) = a*^ v1 +^ a*^ v2.
Proof.
move=> a v1 v2.
by eq_vec.
Qed.


Notation Local " 'Rvec' p " := (vec R p)(at level 20): type_scope.

(*definition and properties for a norm*)
Definition norm (v: vec R p):= \big[Rmax/R0]_(i < p) Rabs (v i).

Lemma norm_elem_eq:
forall (u v:Rvec p),
(forall i, Rabs (u i)=Rabs (v i))-> norm u = norm v.
Proof.
move => u v Huvi;
by apply: eq_bigr=> i _; rewrite Huvi.
Qed.


Lemma elem_le_norm_le:
forall (v1 v2:Rvec p),
(forall i, Rabs (v1 i) <= Rabs (v2 i))-> norm v1 <= norm v2.
Proof.
move=> v1 v2 Hrabs.
apply: big_rel; first by right.
move=> a x b y Hax Hby.
by rewrite /Rmax; case (Rle_dec a b) => Hab; case (Rle_dec x y) => Hxy; 
try fourier; apply Rle_trans with y; try fourier; left; apply Rnot_le_lt .
move=> i _; apply Hrabs.
Qed.


Lemma norm_pos: forall v,R0<=norm v.
Proof.
move => v.
apply: big_prop; [by right|
move=> x y Hx Hy; apply/Rmax_Rle; by left|
move=> i _; apply Rabs_pos].
Qed.


Lemma desc_norm:forall v j, norm v = Rmax (Rabs (v j))
    (\big[Rmax/R0]_(i<p|i!=j) (Rabs (v i))).
Proof.
move=>v j; rewrite /norm.

apply sxbigD1 with (P:=(fun _  :'I_p=> true)) (j:=j) (F:= (fun t => Rabs (v t))).
move => i; apply : Rabs_pos.
done.
Qed.


Lemma norm_max: forall (v:Rvec p) i, Rabs (v i) <= norm v.
Proof.
move=> v i; rewrite (desc_norm v i); apply RmaxLess1.
Qed.

Lemma norm_rew:
(0 < p)%Dn -> forall v,exists k, norm v = Rabs (v k). 
Proof.
move =>H v.
rewrite /norm.
apply: (max_exists ); first by done.
move => i; apply Rabs_pos.
Qed.


Lemma big_ip_zero:
forall T (op: T -> T -> T) nil P n (f: 'I_n -> T) ,
(n = 0)%Dn -> \big[op/nil]_(i<n | P) (f i)  = nil.
move=> T op nil P  n f H.
elim: index_enum.
by rewrite big_nil.
move=> t.
have H1 : (t < n)%Dn by case: t.
have h2: (t < 0)%Dn by rewrite -H.
have H3: ~(t < 0)%Dn by rewrite ltn0.
by elim: H3.
Qed.


Lemma norm_tri: forall (v1 v2:Rvec p), norm (add_v v1 v2)<=norm v1 + norm v2.
Proof.
move=> v1 v2.
have H: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n; by right.
elim: H => H. 
rewrite /norm.
rewrite !big_ip_zero; try done .
right; ring.
elim (norm_rew H (add_v v1 v2))=> k Heq.
rewrite Heq /add_v.
apply Rle_trans with (Rabs (v1 k)+Rabs (v2 k)).
rewrite vxE.
apply Rabs_triang.
apply Rplus_le_compat ; apply norm_max.
Qed.


Lemma  norm_hom_pos:
forall v (a:R),
norm (mult_sv a v) = Rabs a*norm v.
Proof.
move => v a.
have H: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n; by right.
elim: H => H. 
rewrite /norm.
rewrite !big_ip_zero; try done .
ring.

have Hnec: norm (mult_sv a v) <= Rabs a*norm v.
elim (norm_rew H (mult_sv a v)); intros k Heq.
rewrite Heq /mult_sv vxE Rabs_mult.
apply Rmult_le_compat_l; [apply Rabs_pos| apply norm_max].
have H2:Rabs a*norm v<= norm (mult_sv a v).
elim (norm_rew H v); intros k  Heq.
rewrite Heq  -Rabs_mult.
have J: (a* v k)%R = (a*^ v) k.
by rewrite /mult_sv vxE.
rewrite J.
apply (norm_max (mult_sv a v) k ).
by apply Rle_antisym.
Qed.

Lemma Rabs0_R0:
forall r,
Rabs r = R0 <-> r = R0.
Proof.
move => r. 
split =>H.
split_Rabs; apply Rle_antisym; fourier.
rewrite H;apply Rabs_R0.
Qed.


Lemma norm0:
forall v, norm v = R0 <-> v = vect0.
Proof.
move => v.
split; move => H.
rewrite -vecP=> i;  rewrite /vect0 vxE.
have Hineq: Rabs(v i) <= R0.
apply Rle_trans with (norm v).
apply norm_max.
unfold Rle; right; assumption.
have Heq: Rabs (v i)=R0 .
 apply Rle_antisym; [assumption|apply Rabs_pos].
by rewrite -Rabs0_R0.
have H0: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n; by right.
elim: H0 => H0. 
rewrite /norm.
rewrite !big_ip_zero; try done .
elim (norm_rew H0 v)=> k Heq.
rewrite Heq.
rewrite H /vect0 vxE; apply Rabs_R0.
Qed.

Lemma dif_eq_vect0:
forall (u v:Rvec p),
dif_v u v = vect0  <-> u = v.
Proof.
move => u v.
split => H.
rewrite /vect0  in H.
rewrite -vecP => i.
have Hdif: dif_v u v i = R0 by rewrite (H ) vxE.
rewrite /dif_v vxE in Hdif.
by apply Rminus_diag_uniq.
rewrite H.
by eq_vec.
Qed.



Lemma norm0_eq:
forall u v, norm (dif_v u v) = R0 <-> u = v.
Proof.
move => u v.
split=> H.
by move:H; rewrite (norm0 (dif_v u v)) dif_eq_vect0 => H.
by rewrite (norm0 (dif_v u v)) dif_eq_vect0.
Qed.

Lemma Rabs_base:
forall k: 'I_p, Rabs (base_v k k) = R1 .
Proof.
move =>k.
rewrite /base_v vxE.
rewrite eq_refl.
apply Rabs_R1.
Qed.

Lemma norm_base_v:
forall k, norm (base_v k) = R1.
Proof.
move => t.
have H: p = 0%Dn \/ (0 < p)%Dn.
case: p; first by left.
move=> n; by right.
elim: H => H. 

have H1 : (t < p)%Dn by case: t.
have h2: (t < 0)%Dn by rewrite -H.
have H3: ~(t < 0)%Dn by rewrite ltn0.
by elim: H3.

rewrite -(Rabs_base t).

elim (norm_rew H (base_v t)) => k0 Heq.
case C: (t==k0).
by rewrite -(eqP C) in Heq.
rewrite {2}/base_v vxE in Heq.
rewrite C Rabs_R0 in Heq.
have Hco:R1<=R0 by rewrite -Heq -(Rabs_base t); by apply norm_max.
have Hco2: ~(R1<=R0) by fourier.
contradiction.
Qed.


Lemma ex_v_norm:
(0 < p )%Dn -> forall r, (R0 < r)%R -> exists x, norm x = r.
Proof.
move => H r Hpos.
exists (mult_sv r (base_v (I0 H )  )).
rewrite norm_hom_pos norm_base_v Rabs_right.
by ring_simplify.
fourier.
Qed.

Lemma el_div_norm_leq:
forall (v:Rvec p) x, 0<norm v ->Rabs (v x) /norm v <= 1.
Proof.
move => v x Hv.
rewrite /Rdiv.
have H:= norm_max v x.
apply Rmult_le_reg_l with (norm v); first by done.
rewrite Rmult_comm Rmult_assoc Rinv_l; last by apply Rgt_not_eq; apply Hv.
ring_simplify; apply: norm_max.
Qed.



(*definition and properties for distance*)

Definition dist_Rp (x y :Rvec p) :R := norm (dif_v x y).

Lemma dist_pos:
forall v u, R0<= dist_Rp v u.
Proof.
move => v u.
rewrite /dist_Rp.
apply norm_pos.
Qed.

Lemma dist_pos2:
forall v u, dist_Rp v u >= 0. 
Proof.
move => v u; apply Rle_ge;  apply: dist_pos. Qed.

Lemma dist_sym:
forall u v, dist_Rp u v = dist_Rp v u.
Proof.
move=> u v.
rewrite /dist_Rp.
have Heqdif:forall i, Rabs ((dif_v u v) i) = Rabs ((dif_v v u) i).
move => i.
by rewrite /dif_v !vxE Rabs_minus_sym.
by apply norm_elem_eq.
Qed.

Lemma dist_tri:
forall u v t,
dist_Rp u v <= dist_Rp u t + dist_Rp t v.
Proof.
move => u v t.
rewrite /dist_Rp.
have Heq: forall i, (dif_v u v) i = (add_v (dif_v u t) (dif_v t v) )i.
move=>i; rewrite /dif_v /add_v !vxE.
by ring_simplify.
have Hn: norm (dif_v u v) = norm (add_v (dif_v u t) (dif_v t v)).
 apply  norm_elem_eq.
move => i;by  rewrite (Heq i) .
rewrite Hn.
apply norm_tri.
Qed.

Lemma dist_refl:
forall u v,
dist_Rp u v = R0 <-> u = v.
Proof.
move => u v.
 rewrite /dist_Rp.
split => H.
by  rewrite -norm0_eq.
by rewrite norm0_eq.
Qed.

Lemma dif_vect0:
forall v,
dif_v v (@vect0 p) = v. 
Proof.
by move => v; eq_vec.
Qed.


Lemma dist_min_tri:
forall a b c,
dist_Rp a b -dist_Rp c b <= dist_Rp (dif_v a b) (dif_v c b).
Proof.
move=> a b c.
rewrite /dist_Rp.
have H: norm  (a -^ b) <= norm ((a -^ b) -^ (c -^ b)) + norm (c -^ b).
pattern (a -^ b) at 1 ; replace (a -^ b) with (((a -^ b) -^ (c -^ b)) +^ (c -^ b)).
apply norm_tri.
by eq_vec.
fourier.
Qed.


Lemma Vabs_triang:
forall (u v: vec R p), Vabs (u +^ v) <=v: Vabs u +^ Vabs v.
Proof.
move=> u v i; rewrite !vxE; apply Rabs_triang. Qed.

Lemma Vle_trans:
forall (u v w: Rvec p), Vle u v -> Vle v w -> Vle u w.
move=> u v w Huv Hvw i; apply Rle_trans with (v i);
 [apply Huv| apply Hvw].
Qed.

Lemma Vadd_vle_compat:
forall a b c d, Vle a b -> Vle c d -> Vle (a +^ c) (b +^ d).
move=> a b c d Hab Hcd i; rewrite !vxE.
apply Rplus_le_compat; [apply Hab| apply Hcd].
Qed.



End vectors.



Canonical Structure metricSpace_Rp p := 
  Build_Metric_Space  (@dist_pos2 p) 
    (@dist_sym p) (@dist_refl p) (@dist_tri p).
 
Implicit Arguments base_v [p].
Implicit Arguments norm [p].
Implicit Arguments norm_pos [p].
Implicit Arguments norm_hom_pos [p].
Implicit Arguments norm0 [p].
Implicit Arguments dist_Rp [p].
Implicit Arguments dist_pos [p].
Implicit Arguments dist_refl [p].
Implicit Arguments dist_sym [p].
Implicit Arguments dist_tri [p].

Implicit Arguments vec_of_fun [p].
Implicit Arguments norm_rew [p].
Implicit Arguments norm_tri [p].
